from .. import base, experts, project

correct, wrong = base.BinaryEvaluation.get_display_labels()

task_labeling = project.Scenario.EXPERT_LABELING_OF_TASKS.project_name_suffix
solution_labeling = project.Scenario.EXPERT_LABELING_OF_SOLVED_TASKS.project_name_suffix

task_verification = experts.ExpertCase.TASK_VERIFICATION.label
labeling_quality_verification = experts.ExpertCase.LABELING_QUALITY_VERIFICATION.label

image_classification_task = base.LocalizedString(
    {
        'EN': 'classification of animals in photos',
        'RU': 'классификация животных на фотографиях',
    }
)
audio_transcript_task = base.LocalizedString(
    {
        'EN': 'transcript of speech from audio',
        'RU': 'расшифровка речи из аудиозаписей',
    }
)

expert_evaluation = base.LocalizedString(
    {
        'RU': """
<li>Бинарная оценка решения <code>EVAL</code>: <code>{CORRECT}</code>/<code>{WRONG}</code>. Этот пункт всегда должен
отражать, правильно ли в соответствии с инструкцией по аннотации выполнено исходное задание.</li>
    """.format(
            WRONG=wrong['RU'], CORRECT=correct['RU']
        ),
        'EN': """
<li>Binary evaluation of solution <code>EVAL</code>: <code>{CORRECT}</code>/<code>{WRONG}</code>. This option should always
reflect whether the annotation was done accurately according to the instruction or not.</li>
    """.format(
            WRONG=wrong['EN'], CORRECT=correct['EN']
        ),
    }
)

expert_solve_task = base.LocalizedString(
    {
        'RU': '<li>Решить задание, как это сделал бы исполнитель согласно инструкции.</li>',
        'EN': '<li>Complete the task as the worker would do according to the instructions.</li>',
    }
)

expert_solve_annotation_task = base.LocalizedString(
    {
        'RU': """
<li>Решить задание, либо правильно – как это сделал бы исполнитель согласно инструкции, либо специально неправильно –
допустите ошибку, нарушив один из важных пунктов инструкции.</li>
<li>
    В поле <code>EVAL</code> соответственно выберите, правильно выполнили задание или нет.
</li>
""",
        'EN': """
<li>Complete the task either accurately – as the worker would do according to the instructions, or make some errors on
purpose, violating one of the important points of the instructions.</li>
<li>
    In <code>EVAL</code> field choose whether you completed the task accurately or not.
</li>
""",
    }
)

expert_extra_comment_reason = base.LocalizedString(
    {
        'RU': 'допустили при выполнении задания специальную ошибку или',
        'EN': 'made some errors on purpose while completing the task or',
    }
)

expert_fill_eval = base.LocalizedString(
    {
        'RU': """В поле <code>EVAL</code> выберите, верно выполнено задание или нет.""",
        'EN': """In <code>EVAL</code> field choose, whether you've completed the task accurately or not.""",
    }
)

experts = base.LocalizedString(
    {
        'RU': """
<div>
    <h1>Цель задания</h1>

    <p>
        Ваша цель – быть посредником между заказчиком и исполнителями, предоставляя двусторонние гарантии:
    </p>

    <ul>
        <li>Для исполнителей – гарантия предоставления четко поставленной задачи.</li>
        <li>Для заказчика – гарантия предоставления качественных решений.</li>
    </ul>

    <p><strong>Объект</strong> – проверяемое вами задание или задание с решением (далее просто <em>решение</em>).</p>

    <p><strong>Источник объекта</strong> – заказчик или исполнитель.</p>

    <p>
        Объекты, их источники, а также необходимые действия, определяются <strong>сценарием</strong>, указанным в
        карточке выполняемого вами задания.
    </p>

    <h1>Разметка объектов</h1>

    <p>
        Интерфейс предлагаемых вам заданий всегда построен одинаковым образом: сначала следует интерфейс
        <em>объекта</em>, затем фиксированная часть с его экспертной разметкой, отделяемая горизонтальной чертой.
    </p>

    <p>
        Интерфейс <em>объекта</em> зависит от решаемой заказчиком задачи. Например, если он решает задачу
        <em>{TASK_NAME}</em>, то интерфейс целиком может выглядеть так:
    </p>

    <div>
        <input id="hd-30" class="hide" type="checkbox">
        <label for="hd-30"><strong>Пример <em>объекта</em>-задания</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-tasks_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-tasks_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <div>
        <input id="hd-31" class="hide" type="checkbox">
        <label for="hd-31"><strong>Пример <em>объекта</em>-решения</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-solutions_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-solutions_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>
        Разметка <em>объекта</em> включает в себя {EXPERT_LABELING_OBJECT_COUNT} пункта:
    </p>

    <ul>
        {EXPERT_EVALUATION}
        <li>Бинарная оценка <code>OK</code>: <em>{CORRECT}</em>/<em>{WRONG}</em>.</li>
        <li>Текстовый комментарий <code>COMMENT</code>.</li>
    </ul>

    <p>Способ заполнения пунктов <code>OK</code> и <code>COMMENT</code> зависит от <em>сценария</em>.</p>

    <h1>Сценарии</h1>

    <h2>Проверка заданий заказчика</h2>

    <p>
        В ленте заданий Толоки вы увидите карточку с суффиксом <em>({TASK_LABELING})</em> в названии и примечанием <em>{TASK_VERIFICATION}</em>:
    </p>

    <div>
        <input id="hd-32" class="hide" type="checkbox">
        <label for="hd-32"><strong>Пример карточки задания</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-tasks_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-tasks_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>
        Не все заказчики хорошо понимают, как лучше составлять инструкцию и интерфейс заданий так, чтобы исполнители
        четко понимали поставленное задание и не испытывали трудностей при решении.
    </p>

    <p>
        Даже при наличии хорошо составленной инструкции, размечаемые данные могут не соответствовать написанному в
        инструкции.
    </p>

    <p>
        Перечислим критерии хорошо сформулированного задания:
    </p>

    <ul>
        <li><u>Полнота</u> – в инструкции перечислены все случаи, наблюдаемые в потоке размечаемых данных.</li>
        <li><u>Понятность</u> – инструкция написана простым и понятным языком.</li>
        <li><u>Непротиворечивость</u> – в инструкции нет пунктов, которые противоречат друг другу.</li>
        <li><u>Краткость</u> – инструкция содержит минимум информации, нужный для объяснения сути задания.</li>
        <li><u>Удобство</u> – интерфейс заданий удобен и предполагает минимум действий для решения задания.</li>
    </ul>

    <p>
        Вам будут показаны задания заказчика. Цель данного сценария:
    </p>

    <ul>
        <li>
            Дать обратную связь заказчику при необходимости доработать задания. Обратная связь бывает:
            <ul>
                <li><u>По-<em>объектная</em></u>, которую вы укажете при разметке объекта.</li>
                <li><u>Общая</u>, которую вы можете сообщить в чате с заказчиком.</li>
            </ul>
        </li>
        <li>
            Сформировать набор обучающих заданий, который должен соответствовать следующим критериям:
            <ul>
                <li>
                    <u>Полнота</u> – покрытие всех случаев из инструкции, при условии наличия соответствующих
                    примеров.
                </li>
                <li>
                    <u>Содержательность</u> – примерно одинаковое количество информации для каждого случая. Чем
                    сложнее случай, тем больше обучающих примеров может потребоваться для необходимого количества
                    информации.
                </li>
            </ul>
        </li>
    </ul>

    <p>
        Для каждого задания вам требуется:
    </p>

    <ul>
        {EXPERT_SOLVE_TASK}
        <li>
            Если задание <strong>некорректное</strong> – соответствующий случай не описан в инструкции, оно вообще
            не соответствует инструкции, возникли технические сложности вроде ошибки 404, либо по какой-то иной
            причине, выберите <code>{WRONG}</code> в поле <code>OK</code> и укажите <code>COMMENT</code> с описанием
            проблемы.
        </li>
        <li>
            Если вышеперечисленных проблем нет, то задание <strong>корректное</strong>, выберите
            <code>{CORRECT}</code> в поле <code>OK</code>. Если вы {EXTRA_COMMENT_REASON} считаете, что задание должно попасть в обучение, укажите <code>COMMENT</code>,
            исполнители с ним смогут ознакомиться при прохождении обучения. В противном случае, оставьте
            <code>COMMENT</code> пустым.
        </li>
    </ul>

    <h2>Проверка решенных заданий заказчика</h2>

    <p>
        В ленте заданий Толоки вы увидите карточку с суффиксом <em>{EXTRA_SUFFIX}({SOLUTION_LABELING})</em> в названии и примечанием <em>{TASK_VERIFICATION}</em>:
    </p>
    <div>
        <input id="hd-33" class="hide" type="checkbox">
        <label for="hd-33"><strong>Пример карточки задания</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-solutions_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-solutions_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>
        Сценарий эквивалентен предыдущему, с тем лишь отличием, что заказчик предоставляет задания с решениями. {EXPERT_FILL_EVAL}
    </p>
    <p>
        Правила разметки <em>объектов</em> (решений) немного изменятся. Решение <em>некорректное</em>, если:
    </p>

    <ul>
        <li>Некорректно задание (см. выше критерии и соответствующие действия)</li>
        <li>
            Некорректно решение – оно не соответствует инструкции, выберите <code>{WRONG}</code> в поле <code>OK</code> и укажите
            причину в <code>COMMENT</code>.
        </li>
    </ul>

    <h2>Проверка качества разметки</h2>

    <p>
        В ленте заданий Толоки вы увидите карточку с суффиксом <em>({SOLUTION_LABELING})</em> в названии и примечанием <em>{LABELING_QUALITY_VERIFICATION}</em>:
    </p>

    <div>
        <input id="hd-34" class="hide" type="checkbox">
        <label for="hd-34"><strong>Пример карточки задания</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_verification_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="verification-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_verification_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="verification-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>Сценарий эквивалентен предыдущему. Источник решений в данном сценарии – разметка исполнителей. Оценивать нужно только правильность решения. </p>

    <p>
        Решение <em>некорректное</em>, если оно не соответствует инструкции, выберите <code>{WRONG}</code> в поле
        <code>OK</code> и опционально укажите причину в <code>COMMENT</code>.
    </p>
</div>
""",
        'EN': """
<div>
    <h1>Task purpose</h1>

    <p>
        Your goal is to be an intermediary between the customer and the workers, providing bilateral guarantees:
    </p>

    <ul>
        <li>For performers, it is a guarantee of providing a clearly defined task.</li>
        <li>For the customer – a guarantee of providing quality solutions.</li>
    </ul>

    <p><strong>Object</strong> – the task or the task with the solution (next, just a <em>solution</em>) you are checking.</p>

    <p><strong>Object source</strong> – customer or worker.</p>

    <p>
        The objects, their sources, as well as the necessary actions are determined by the <strong>scenario</strong>
        specified in the card of the task you are performing.
    </p>

    <h1>Objects labeling</h1>

    <p>
        The interface of the tasks offered to you is always built in the same way: the interface follows first of the
        <em>object</em>, then the fixed part with its expert labeling, separated by a horizontal line.
    </p>

    <p>
        The interface of the <em>object</em> depends on the task being solved by the customer. For example, if he
        solves a problem <em>{TASK_NAME}</em>, then the entire interface may look like this:
    </p>

    <div>
        <input id="hd-30" class="hide" type="checkbox">
        <label for="hd-30"><strong>Task <em>object</em> example</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-tasks_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-tasks_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <div>
        <input id="hd-31" class="hide" type="checkbox">
        <label for="hd-31"><strong>Solution <em>object</em> example</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-solutions_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-ui_expert-labeling-of-solutions_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-object" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>
        <em>Object</em> labeling includes {EXPERT_LABELING_OBJECT_COUNT} items:
    </p>

    <ul>
        {EXPERT_EVALUATION}
        <li>Binary evaluation <code>OK</code>: <em>{CORRECT}</em>/<em>{WRONG}</em>.</li>
        <li>Textual comment <code>COMMENT</code>.</li>
    </ul>

    <p>The way to fill in these items depends on the <em>scenario</em>.</p>

    <h1>Scenarios</h1>

    <h2>Checking the customer's tasks</h2>

    <p>
        In the Toloka task feed, you will see a card with the suffix <em>({TASK_LABELING})</em> in the title and the note <em>{TASK_VERIFICATION}</em>:
    </p>

    <div>
        <input id="hd-32" class="hide" type="checkbox">
        <label for="hd-32"><strong>Task card example</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-tasks_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-tasks_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>
        Not all customers understand well how best to compile instructions and the interface of tasks so that workers
        clearly understand the task and had no difficulties in performing it.
    </p>

    <p>
        Let's list the criteria for a well-formulated task:
    </p>

    <ul>
        <li><u>Completeness</u> - the instructions list all the cases observed in the data stream.</li>
        <li><u>Clarity</u> - the instruction is written in a simple and understandable language.</li>
        <li><u>Consistency</u> – there are no points in the instructions that contradict each other.</li>
        <li><u>Brevity</u> - the instruction contains a minimum of information necessary to explain the essence of the
            task.</li>
        <li><u>Convenience</u> - the task interface is convenient and involves a minimum of actions to perform the
            task.</li>
    </ul>

    <p>
        You will be shown the customer's tasks. The purpose of this scenario:
    </p>

    <ul>
        <li>
            Give feedback to the customer, if necessary, to rework the tasks. Types of feedback are:
            <ul>
                <li><u>By-<em>object</em></u>, which you specify when labeling the object.</li>
                <li><u>General</u>, which you can inform in a chat with the customer.</li>
            </ul>
        </li>
        <li>
            Create a set of training tasks that must meet the following criteria:
            <ul>
                <li>
                    <u>Completeness</u> - coverage of all cases from the instructions, subject to the availability of
                    appropriate examples.
                </li>
                <li>
                    <u>Informativeness</u> – approximately the same amount of information for each case. The more
                    complicated the case, the more training examples may be required for the required amount of
                    information.
                </li>
            </ul>
        </li>
    </ul>

    <p>
        For each task you need:
    </p>

    <ul>
        {EXPERT_SOLVE_TASK}
        <li>Perform the task as the worker would do according to the instructions.</li>
        <li>
            If the task is <strong>incorrect</strong> - the corresponding case is not described in the instructions, it
            is generally does not comply with the instructions, there were technical difficulties like error 404, or
            for some other reason, choose <code>{WRONG}</code> in <code>OK</code> field and specify
            <code>COMMENT</code> with a problems description.
        </li>
        <li>
            If there are no problems listed above, then the task is <strong>correct</strong>, choose
            <code>{CORRECT}</code> in <code>OK</code> field. If you {EXTRA_COMMENT_REASON} think that the task should be
            included in the training, specify <code>COMMENT</code>, which workers will be able to learn during
            training. Otherwise, leave <code>COMMENT</code> empty.
        </li>
    </ul>

    <h2>Checking the customer's solved tasks</h2>

    <p>
        In the Toloka task feed, you will see a card with the suffix <em>{EXTRA_SUFFIX}({SOLUTION_LABELING})</em> in the title and the note <em>{TASK_VERIFICATION}</em>:
    </p>

    <div>
        <input id="hd-33" class="hide" type="checkbox">
        <label for="hd-33"><strong>Task card example</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-solutions_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_test-launch-on-solutions_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>The scenario is equivalent to the previous one, with the only difference that the customer provides tasks with
        solutions. {EXPERT_FILL_EVAL}</p>

    <p>
        The rules for labeling of <em>objects</em> (solutions) will change slightly. The solution is
        <em>incorrect</em> if:
    </p>

    <ul>
        <li>The task is incorrect (see the criteria and corresponding actions above)</li>
        <li>
            The solution is incorrect – it does not comply with the instructions, choose <code>{WRONG}</code> in
            <code>OK</code> field and specify the reason in <code>COMMENT</code>.
        </li>
    </ul>
    <h2>Labeling quality verification</h2>

    <p>
        In the Toloka task feed, you will see a card with the suffix <em>({SOLUTION_LABELING})</em> in the title and the note <em>{LABELING_QUALITY_VERIFICATION}</em>:
    </p>

    <div>
        <input id="hd-34" class="hide" type="checkbox">
        <label for="hd-34"><strong>Task card example</strong></label>
        <div>
            <table>
                <tbody>
                    <tr>
                        <td style="text-align:center"><h2>Browser</h2></td>
                        <td style="text-align:center"><h2>Mobile app</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_verification_{ANNOTATION_URL_SUFFIX}browser.png"
                                 alt="task-card-browser" style="width:100%;max-width:500px"/>
                        </td>
                        <td>
                            <img src="https://storage.yandexcloud.net/dataforge-public/media/instructions/experts/en/task-card_verification_{ANNOTATION_URL_SUFFIX}app.png"
                                 alt="task-card-app" style="width:100%;max-width:500px"/>
                        </td>
                    </tr>
                </tbody>
            </table>
        </div>
    </div>

    <p>The scenario is equivalent to the previous one. In this scenario tasks were solved by workers. You should only evaluate the correctness of the <strong>solution</strong></p>

    <p>
        The solution is <em>incorrect</em> if it does not comply with the instructions, choose <code>{WRONG}</code>
            in <code>OK</code> field and specify the reason in <code>COMMENT</code>.
    </p>
</div>
""",
    }
)

worker_header = base.LocalizedString(
    {
        'RU': '<h1>Инструкция для исполнителей</h1>',
        'EN': '<h1>Instructions for workers</h1>',
    }
)

expert_cut = base.LocalizedString(
    {
        'RU': """
<input id="hd-101" class="hide" type="checkbox">
<label for="hd-101"><strong>Экспертная разметка</strong></label>
""",
        'EN': """
<input id="hd-101" class="hide" type="checkbox">
<label for="hd-101"><strong>Expert labeling</strong></label>
""",
    }
)


def get_formatted_expert_instruction(task_function: base.TaskFunction, lang: str, has_evaluation: bool):
    is_annotation = isinstance(task_function, base.AnnotationFunction)
    return experts[lang].format(
        EXPERT_LABELING_OBJECT_COUNT=2 if not has_evaluation else 3,
        EXPERT_EVALUATION=expert_evaluation[lang] if has_evaluation else '',
        EXPERT_SOLVE_TASK=(expert_solve_task if not is_annotation else expert_solve_annotation_task)[lang],
        EXTRA_COMMENT_REASON=expert_extra_comment_reason[lang] if is_annotation else '',
        EXTRA_SUFFIX=f'{project.CHECK[lang]} ' if is_annotation else '',
        EXPERT_FILL_EVAL=expert_fill_eval[lang] if has_evaluation else '',
        TASK_LABELING=task_labeling[lang],
        SOLUTION_LABELING=solution_labeling[lang],
        TASK_VERIFICATION=task_verification[lang],
        LABELING_QUALITY_VERIFICATION=labeling_quality_verification[lang],
        TASK_NAME=audio_transcript_task[lang] if is_annotation else image_classification_task[lang],
        ANNOTATION_URL_SUFFIX='annotation_' if is_annotation else '',
        WRONG=wrong[lang],
        CORRECT=correct[lang],
    )



from copy import copy
from dataclasses import dataclass, field
from datetime import timedelta
from typing import List, Dict, Tuple, Callable

from IPython.display import display
import ipywidgets as widgets
import pandas as pd
import toloka.client as toloka

from .. import client

from . import params
from .base import ParamName, Update, UIEvent, SpeedControlChangeType, SpeedControlUpdate, SpeedControlEvent
from .event_loop import EventLoop, get_loop, AnnotationEventLoop, get_loop_for_annotation
from .util import get_rec_label
from .characteristics import _get_cumulative_stats_and_characteristics


def get_inner_widget(param: params.Param) -> widgets.Widget:
    if isinstance(param, params.Overlap):
        return widgets.Combobox(
            value=param.value,
            placeholder='2-3',
            options=param.available_range,
            ensure_option=False,
            disabled=param.disabled,
            continuous_update=False,
        )

    obj = param.available_range[0]
    if isinstance(obj, str):
        widget = widgets.Dropdown(
            options=param.available_range,
            disabled=param.disabled,
            value=param.value,
        )
    elif isinstance(obj, bool):
        widget = widgets.Checkbox(
            disabled=param.disabled,
            value=param.value,
        )
    elif isinstance(obj, int):
        if isinstance(param.value, timedelta):
            value = int(param.value.total_seconds())
        else:
            value = param.value
        widget = widgets.BoundedIntText(
            min=param.available_range[0],
            max=param.available_range[-1],
            step=1,
            disabled=param.disabled,
            value=value,
        )
    elif isinstance(obj, float):
        widget = widgets.BoundedFloatText(
            min=param.available_range[0],
            max=param.available_range[-1],
            step=0.01,
            disabled=param.disabled,
            value=param.value,
            readout=True,
            readout_format='.2f',
        )
    else:
        raise NotImplementedError(f'{obj}')
    return widget


def get_speed_control_widget(
    param: params.SpeedControlList, loop: EventLoop, update_callback: Callable
) -> widgets.Widget:
    grid = widgets.GridspecLayout(param.list_size + 1, 1, align_items='center')

    fmt = widgets.HTML('<h5>Format: 1d 21h 35m 60s</h5>', layout={'width': '300px'})
    ratio = widgets.HTML('<h5>Ratio threshold</h5>', layout={'width': '294px'})
    reject = widgets.HTML('<h5>Reject?</h5>')
    help_button = widgets.Button(description='?', layout={'width': '30px'}, button_style='info')

    def on_button_clicked(button):
        loop.start(
            UIEvent(
                sender=ParamName.HELP.value,
                update={
                    ParamName.HELP: Update(f'<b>Help for {param.description}</b>: {param.help_msg}', button.disabled)
                },
            )
        )
        if loop.ui_initialized:
            update_callback()

    help_button.on_click(on_button_clicked)
    header = widgets.Box([fmt, ratio, reject, help_button])
    grid[0, :] = header

    def get_row(index: int, value: params.BlockTimePicker = copy(params.default_block)) -> widgets.Widget:
        time_widget = widgets.Text(
            value=value.time_value,
            continuous_update=False,
        )
        reject = widgets.Checkbox(
            value=value.reject,
            indent=False,
            layout=widgets.Layout(height='auto', width='40px', padding='0px 0px 0px 10px'),
        )
        ratio = widgets.BoundedFloatText(
            min=0.0,
            max=1.0,
            step=0.01,
            value=value.ratio,
        )

        remove_button = widgets.Button(
            icon='trash',
            button_style='danger',
            layout={'width': '30px'},
        )

        def on_time_change(change):
            s = change['new'].replace('y', 'Y')
            try:
                td = pd.Timedelta(s)
                assert param.min_value <= td.total_seconds() <= param.max_value
                loop.start(
                    params.SpeedControlEvent(
                        param.name.ui,
                        {
                            SpeedControlChangeType.TIME: SpeedControlUpdate(
                                change['new'], param.disabled, SpeedControlChangeType.TIME, index
                            )
                        },
                    )
                )
                if loop.ui_initialized:
                    update_callback()
            except:
                change['owner'].value = change['old']

        def on_ratio_change(change):
            loop.start(
                SpeedControlEvent(
                    param.name.ui,
                    {
                        SpeedControlChangeType.RATIO: SpeedControlUpdate(
                            change['new'], param.disabled, SpeedControlChangeType.RATIO, index
                        )
                    },
                )
            )
            if loop.ui_initialized:
                update_callback()

        def on_reject_change(change):
            loop.start(
                SpeedControlEvent(
                    param.name.ui,
                    {
                        SpeedControlChangeType.REJECT: SpeedControlUpdate(
                            change['new'], param.disabled, SpeedControlChangeType.REJECT, index
                        )
                    },
                )
            )
            if loop.ui_initialized:
                update_callback()

        def on_remove_click(button):
            loop.start(
                SpeedControlEvent(
                    param.name.ui,
                    {
                        SpeedControlChangeType.REMOVE: SpeedControlUpdate(
                            '', param.disabled, SpeedControlChangeType.REMOVE, index
                        )
                    },
                )
            )
            if loop.ui_initialized:
                update_callback()

        time_widget.observe(on_time_change, 'value')
        ratio.observe(on_ratio_change, 'value')
        reject.observe(on_reject_change, 'value')
        remove_button.on_click(on_remove_click)
        return widgets.Box([time_widget, ratio, reject, remove_button])

    grid.reserved_rows = [get_row(i) for i in range(param.list_size)]

    grid.add_button = widgets.Button(
        description='+',
        button_style='success',
        layout={'width': '686px'},
    )

    def on_add_click(button):
        loop.start(
            SpeedControlEvent(
                param.name.ui,
                {
                    SpeedControlChangeType.ADD: SpeedControlUpdate(
                        '',
                        param.disabled,
                        SpeedControlChangeType.ADD,
                    )
                },
            )
        )
        if loop.ui_initialized:
            update_callback()

    grid.add_button.on_click(on_add_click)
    return grid


def get_default_widget(param: params.Param, loop: EventLoop, update_callback: Callable) -> widgets.Widget:
    obj = param.available_range[0]
    widget = get_inner_widget(param)
    desc = widgets.HTML(value=f"<font color='black'>{param.description}</font>")
    help_button = widgets.Button(description='?', layout={'width': '30px'}, button_style='info')

    def on_button_clicked(button):
        loop.start(
            UIEvent(
                sender=ParamName.HELP.value,
                update={
                    ParamName.HELP: Update(f'<b>Help for {param.description}</b>: {param.help_msg}', button.disabled)
                },
            )
        )
        if loop.ui_initialized:
            update_callback()

    help_button.on_click(on_button_clicked)
    rec = widgets.HTML(value=f"<font color='black'>{get_rec_label(obj, param.recommended_range)}</font>")
    outer_widget = widgets.Box(
        [widgets.Box([desc, help_button], layout={'justify_content': 'space-between'}), widget, rec]
    )

    def on_change(change):
        loop.start(UIEvent(param.name.ui, {param.name: Update(change['new'], param.disabled)}))
        if loop.ui_initialized:
            update_callback()

    widget.observe(on_change, 'value')
    return outer_widget


def get_widget(param: params.Param, loop: EventLoop, update_callback: Callable) -> widgets.Widget:
    create_function = {
        params.Help: lambda *_: widgets.HTML('', layout={'height': '60px'}),
        params.Stat: lambda *_: widgets.HBox([widgets.Label(value=param.description), widgets.IntProgress()]),
        params.StatWithExtraText: lambda *_: widgets.HBox(
            [widgets.Label(value=param.description), widgets.IntProgress(), widgets.HTMLMath('')]
        ),
        params.Characteristics: lambda *_: widgets.HTMLMath(''),
        params.SpeedControlList: get_speed_control_widget,
    }.get(type(param), get_default_widget)
    w = create_function(param, loop, update_callback)
    dispatch(param, w)
    return w


def dispatch_default(param: params.Param, widget: widgets.Widget):
    w = widget.children[1]
    obj = param.available_range[0]
    color = 'black'
    if isinstance(obj, str) or isinstance(obj, bool):
        if param.value not in param.recommended_range:
            color = 'red'
    else:
        w.min = param.available_range[0]
        w.max = param.available_range[-1]

        if not (param.recommended_range[0] <= param.value <= param.recommended_range[-1]):
            color = 'red'

    widget.children[-1].value = f"<font color='{color}'>{get_rec_label(obj, param.recommended_range)}</font>"
    w.value, w.disabled = param.value, param.disabled


def dispatch_assignment_price(param: params.AssignmentPrice, widget: widgets.Widget):
    dispatch_default(param, widget)
    # if min bound for widget is 0.005, with step 0.01, values like 1.25 are "incorrect",
    # because they cannot be reached by ipywidgets logic, which contradicts toloka's logic with dynamic price step

    # however, there is still one edge effect - if param.value=1$, steps for widget increase and decrease
    # logically should be different. this cannot be done, as widget can have only one step value
    # so, current behavior is:
    # 0.9998 -> 0.9999 -> 1.0 -> [1.001 -> automatically updates in Param logic to] 1.01
    # and 1.01 -> 1.0 -> 0.999
    widget.children[1].step = params.get_price_step(param.value)


def dispatch_help(param: params.Help, widget: widgets.Widget):
    widget.value = param.value


def dispatch_stat(param: params.Stat, widget: widgets.Widget):
    v = param.value
    v = max(0.0, v)
    v = min(1.0, v)

    widget.children[1].value = v * 100

    b = 0
    if v < 0.5:
        r = 255
        g = int(2 * 255 * v)
    else:
        v = min(v, 1.0)
        g = 255
        r = int(255 * (1.0 - 2 * (v - 0.5)))

    if param.reverse:
        r, g = g, r

    widget.children[1].style.bar_color = '#%02x%02x%02x' % (r, g, b)


def dispatch_stat_with_extra_text(param: params.StatWithExtraText, widget: widgets.Widget):
    value, text = param.value
    stat = params.Stat(value=value, description=None, name=None)
    dispatch_stat(stat, widget)
    widget.children[-1].value = text


def dispatch_characteristics(param: params.Characteristics, widget: widgets.Widget):
    widget.value = '<br>'.join(param.value)


def dispatch_task_duration(param: params.TaskDuration, widget: widgets.Widget):
    widget.value = int(param.value.total_seconds())


def dispatch_speed_control(param: params.SpeedControlList, widget: widgets.Widget):
    widget.add_button.disabled = param.disabled
    for i, (row, item) in enumerate(zip(widget.reserved_rows, param.value)):
        row.children[0].value, row.children[0].disabled = item.time_value, param.disabled
        row.children[1].value, row.children[1].disabled = item.ratio, param.disabled
        row.children[2].value, row.children[2].disabled = item.reject, param.disabled
        row.children[3].disabled = param.disabled
        widget[i + 1, :] = row

    if len(param.value) < param.list_size:
        widget[len(param.value) + 1, :] = widget.add_button

    for i in range(len(param.value) + 2, param.list_size + 1):
        widget[i, :] = widgets.Label()


def dispatch(param: params.Param, widget: widgets.Widget):
    dispatch_function = {
        params.Help: dispatch_help,
        params.Stat: dispatch_stat,
        params.StatWithExtraText: dispatch_stat_with_extra_text,
        params.Characteristics: dispatch_characteristics,
        params.TaskDuration: dispatch_task_duration,
        params.SpeedControlList: dispatch_speed_control,
        params.AssignmentPrice: dispatch_assignment_price,
        params.AnnotationAssignmentPrice: dispatch_assignment_price,
    }.get(type(param), dispatch_default)
    dispatch_function(param, widget)


@dataclass
class UI:
    loop: EventLoop
    ui: widgets.Widget = field(init=False)
    param_to_widget: Dict[ParamName, widgets.Widget] = field(init=False)

    def __post_init__(self):
        self.param_to_widget = {
            p.name: get_widget(p, self.loop, self.update_widgets)
            for p in self.loop.params + self.loop.stats + [self.loop.characteristics]
            if isinstance(p, params.Param)
        }
        self.ui = create_ui(self.param_to_widget)
        self.loop.ui_initialized = True

    def update_widgets(self):
        for param in self.loop.params + self.loop.stats + [self.loop.characteristics]:
            if isinstance(param, params.Param):
                dispatch(param, self.param_to_widget[param.name])

    def get_params(self) -> client.Params:
        return self.loop.get_params()


@dataclass
class UIAnnotation:
    check_loop: EventLoop
    markup_loop: AnnotationEventLoop
    ui: widgets.Widget = field(init=False)
    check_param_to_widget: Dict[ParamName, widgets.Widget] = field(init=False)
    markup_param_to_widget: Dict[ParamName, widgets.Widget] = field(init=False)
    cumulative_param_to_widget: Dict[ParamName, widgets.Widget] = field(init=False)
    cumulative_stats: List[params.Stat] = field(init=False)
    cumulative_characteristics: params.Characteristics = field(init=False)

    def __post_init__(self):

        self.check_param_to_widget = {
            p.name: get_widget(p, self.check_loop, self.update_widgets)
            for p in self.check_loop.params + self.check_loop.stats + [self.check_loop.characteristics]
            if isinstance(p, params.Param)
        }
        self.markup_param_to_widget = {
            p.name: get_widget(p, self.markup_loop, self.update_widgets)
            for p in self.markup_loop.params + self.markup_loop.stats + [self.markup_loop.characteristics]
            if isinstance(p, params.Param)
        }
        check_ui = create_ui(self.check_param_to_widget)
        markup_ui = create_markup_ui(self.markup_param_to_widget)
        ui = widgets.Accordion(children=[markup_ui, check_ui])
        ui.set_title(0, 'Annotation')
        ui.set_title(1, 'Check')

        ui.selected_index = 0
        cumulative_quality = params.StatWithExtraText(ParamName.QUALITY, 'Quality')
        cumulative_time = params.Stat(ParamName.TIME, 'Time', reverse=True)
        cumulative_cost = params.Stat(ParamName.COST, 'Cost', reverse=True)
        self.cumulative_stats = [cumulative_quality, cumulative_time, cumulative_cost]
        self.cumulative_characteristics = params.Characteristics()
        self.cumulative_param_to_widget = {
            p.name: get_widget(p, None, lambda: None) for p in self.cumulative_stats + [self.cumulative_characteristics]
        }
        cumulative_stats_w = get_grid([self.cumulative_param_to_widget[p.name] for p in self.cumulative_stats])

        cumulative_characteristics_w = self.cumulative_param_to_widget[self.cumulative_characteristics.name]

        self.ui = widgets.VBox([ui, cumulative_stats_w, cumulative_characteristics_w])
        self.update_widgets()
        self.check_loop.ui_initialized = True
        self.markup_loop.ui_initialized = True

    def update_widgets(self):
        for loop, param_to_widget in [
            (self.check_loop, self.check_param_to_widget),
            (self.markup_loop, self.markup_param_to_widget),
        ]:
            for param in loop.params + loop.stats + [loop.characteristics]:
                if isinstance(param, params.Param):
                    dispatch(param, param_to_widget[param.name])

        check_cumulative_params = {
            **{
                'check_' + name.value: self.check_loop.state[name].value
                for name in [
                    ParamName.TASK_COUNT,
                    ParamName.CONTROL_TASK_COUNT,
                    ParamName.ASSIGNMENT_PRICE,
                    ParamName.OVERLAP,
                    ParamName.OVERLAP_CONFIDENCE,
                ]
            },
            **{'check_' + p.name.value: p.value for p in self.check_loop.stats},
        }
        markup_cumulative_params = {
            **{
                'markup_' + name.value: self.markup_loop.state[name].value
                for name in [
                    ParamName.TASK_COUNT,
                    ParamName.ASSIGNMENT_PRICE,
                    ParamName.OVERLAP,
                    ParamName.OVERLAP_CONFIDENCE,
                    ParamName.CHECK_SAMPLE,
                    ParamName.MAX_TASKS_TO_CHECK,
                    ParamName.ACCURACY_THRESHOLD,
                ]
            },
            **{'markup_' + p.name.value: p.value for p in self.markup_loop.stats},
        }
        stats, characteristics = _get_cumulative_stats_and_characteristics(
            **check_cumulative_params, **markup_cumulative_params
        )
        for p, v in zip(self.cumulative_stats, stats):
            p.value = v
            dispatch(p, self.cumulative_param_to_widget[p.name])

        self.cumulative_characteristics.value = characteristics
        dispatch(self.cumulative_characteristics, self.cumulative_param_to_widget[self.cumulative_characteristics.name])

    def get_params(self) -> Tuple[client.Params, client.AnnotationParams]:
        return self.check_loop.get_params(), self.markup_loop.get_params()


def get_grid(ws: List[widgets.Widget]) -> widgets.Widget:
    grid = widgets.GridspecLayout(len(ws), 3)
    for i, w in enumerate(ws):
        try:
            for j, c in enumerate(w.children):
                grid[i, j] = c
        except:
            grid[i, :] = w
    return grid


def create_ui(ws: Dict[ParamName, widgets.Widget]) -> widgets.Widget:
    general_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.TASK_DURATION_HINT,
                ParamName.LANGUAGE,
                ParamName.TASK_COUNT,
                ParamName.ASSIGNMENT_PRICE,
            ]
        ]
    )

    control_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.QUALITY_PRESET,
                ParamName.CONTROL_TASK_COUNT,
                ParamName.CONTROL_TASKS_FOR_ACCEPT,
                ParamName.CONTROL_TASKS_FOR_BLOCK,
                ParamName.OVERLAP,
                ParamName.OVERLAP_CONFIDENCE,
                ParamName.AGGREGATION_ALGORITHM,
            ]
        ]
    )
    qual_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.VERIFIED_LANG,
                ParamName.AGE_RESTRICTION,
                ParamName.TRAINING,
                ParamName.TRAINING_SCORE,
            ]
        ]
    )
    speed_control = ws[ParamName.SPEED_CONTROL]

    tab = widgets.Tab()
    tab.children = [general_grid, qual_grid, control_grid, speed_control]
    titles = ('General', 'Qualifications', 'Control', 'Speed control')

    for i, title in enumerate(titles):
        tab.set_title(i, title)

    tab.layout.height = '348px'

    stats = get_grid([ws[name] for name in [ParamName.QUALITY, ParamName.TIME, ParamName.COST]])
    chars = ws[ParamName.CHARACTERISTICS]
    ui = widgets.VBox([ws[ParamName.HELP], tab, stats, chars])
    return ui


def create_markup_ui(ws: Dict[ParamName, widgets.Widget]) -> widgets.Widget:
    general_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.TASK_DURATION_HINT,
                ParamName.LANGUAGE,
                ParamName.TASK_COUNT,
                ParamName.ASSIGNMENT_PRICE,
            ]
        ]
    )

    control_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.QUALITY_PRESET,
                ParamName.CONTROL_TASKS_FOR_ACCEPT,
                ParamName.CONTROL_TASKS_FOR_BLOCK,
                ParamName.OVERLAP,
                ParamName.OVERLAP_CONFIDENCE,
            ]
        ]
    )
    check_sample_grid = get_grid(
        [
            ws[name]
            for name in [
                ParamName.CHECK_SAMPLE,
                ParamName.MAX_TASKS_TO_CHECK,
                ParamName.ACCURACY_THRESHOLD,
            ]
        ]
    )
    qual_grid = get_grid([ws[name] for name in [ParamName.VERIFIED_LANG, ParamName.AGE_RESTRICTION]])
    speed_control = ws[ParamName.SPEED_CONTROL]

    tab = widgets.Tab()
    tab.children = [general_grid, qual_grid, control_grid, check_sample_grid, speed_control]
    titles = ('General', 'Qualifications', 'Control', 'Check sample', 'Speed control')

    for i, title in enumerate(titles):
        tab.set_title(i, title)

    tab.layout.height = '275px'

    stats = get_grid([ws[name] for name in [ParamName.QUALITY, ParamName.TIME, ParamName.COST]])
    chars = ws[ParamName.CHARACTERISTICS]
    ui = widgets.VBox([ws[ParamName.HELP], tab, stats, chars])
    return ui


def get_interface(
    task_spec: client.PreparedTaskSpec,
    task_duration_hint: timedelta,
    toloka_client: toloka.TolokaClient,
) -> UI:
    loop = get_loop(task_spec, task_duration_hint, toloka_client)
    ui = UI(loop)
    display(ui.ui)
    return ui


def get_annotation_interface(
    task_spec: client.AnnotationTaskSpec,
    check_task_duration_hint: timedelta,
    annotation_task_duration_hint: timedelta,
    toloka_client: toloka.TolokaClient,
) -> UIAnnotation:
    check_loop = get_loop(task_spec.check, check_task_duration_hint, toloka_client)
    markup_loop = get_loop_for_annotation(task_spec, annotation_task_duration_hint, toloka_client)
    ui = UIAnnotation(check_loop, markup_loop)
    display(ui.ui)
    return ui



from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Type, TypeVar, Generic, Optional, Tuple, Any, Set

import toloka.client.project.template_builder as tb

from .common import TextFormat, Title
from .conditions import Consequence, If
from .types import Label, Object, Class, Evaluation, SbSChoice, ImageAnnotation

T = TypeVar('T', covariant=True)
ObjectT = TypeVar('ObjectT', bound=Type[Object], covariant=True)
ClassT = TypeVar('ClassT', bound=Type[Class], covariant=True)
EvaluationT = TypeVar('EvaluationT', bound=Type[Evaluation], covariant=True)
SbSChoiceT = TypeVar('SbSChoiceT', bound=Type[SbSChoice], covariant=True)
ImageAnnotationT = TypeVar('ImageAnnotationT', bound=Type[ImageAnnotation], covariant=True)


@dataclass
class ObjectMeta(Generic[ObjectT]):
    type: ObjectT
    name: Optional[str] = None
    title: Optional[Title] = None
    required: bool = True

    def replace_name(self, name: str):
        copy = deepcopy(self)
        copy.name = name
        return copy


@dataclass
class AvailableLabels(Consequence):
    labels: List[Label]


@dataclass
class SingleLabel(Consequence):
    label: Label


def create_available_labels_if(name: str, value_labels: List[Tuple[Any, List[Class]]]) -> If:
    return If.generate_list_sequence(name, [(value, AvailableLabels(labels)) for value, labels in value_labels])


def create_input_label_as_text_if(name: str, labels: List[Class]) -> If:
    return If.generate_list_sequence(name, [(label.value, SingleLabel(label)) for label in labels])


class LabelsDisplayType(Enum):
    """
    Way to display label set as input or output argument.

    MULTI: for input - display as radio group button
           for output - display as disabled radio group button with chosen label being selected

    MONO: for input - display as text area with chosen label
          for output - display as dropdown list or text area with auto-suggest (not implemented)

    In any case, if label set size is big enough, it displays as for MONO.
    """

    MULTI = 'multi'
    MONO = 'mono'


@dataclass
class ClassMeta(ObjectMeta[ClassT]):
    available_labels: Optional[If] = None
    input_display_type: LabelsDisplayType = LabelsDisplayType.MULTI
    text_format: TextFormat = TextFormat.PLAIN


@dataclass
class ImageAnnotationMeta(ObjectMeta[ImageAnnotationT]):
    available_shapes: Set[tb.fields.ImageAnnotationFieldV1.Shape] = field(
        default_factory=set(tb.fields.ImageAnnotationFieldV1.Shape)
    )
    labels: Optional[Type[Class]] = None


class EvaluationMeta(ObjectMeta[EvaluationT]):
    ...


class SbSChoiceMeta(ObjectMeta[SbSChoiceT]):
    ...



from copy import copy
from dataclasses import dataclass, field
from datetime import timedelta
from math import ceil
from typing import List, Tuple, Optional

import pandas as pd
import toloka.client as toloka
from toloka.client.filter import Languages


from .. import client, experts
from ..pricing import (
    _get_assignment_params,
    _get_random_chance,
    min_assignment_cost,
    min_tasks_in_assignment,
    max_tasks_in_assignment,
    threshold_for_assignment_cost_granularity as threshold_cost,
    round_price,
)
from ..control import BlockTimePicker

from .base import (
    OptionType,
    Node,
    ParamName,
    Update,
    Event,
    UIEvent,
    SpeedControlEvent,
    SpeedControlChangeType,
    QualityPresetType,
)
from .util import get_overlap, overlap_is_static


default_block = BlockTimePicker(0.05, '30m', False)


@dataclass
class Param(Node):
    description: str
    value: OptionType
    available_range: List[OptionType]
    recommended_range: List[OptionType]
    help_msg: str
    disabled: bool = False

    def process(self, event: Event) -> List[Event]:
        if self.name in event.update:
            self.value, self.disabled = event.update[self.name].value, event.update[self.name].disabled
        return []


@dataclass
class ParamGroup(Node):
    params: List[Param]

    def recalculate(self, event: Event) -> Event:
        return event

    def process(self, event: Event) -> List[Event]:
        update = event.update

        for p in self.params:
            if p.name not in update:
                update[p.name] = Update(p.value, p.disabled)

        event = Event(
            sender=self.name.value,
            update=update,
        )
        return [self.recalculate(event)]


@dataclass
class ParamGroupControlTasksRecalculate(ParamGroup):
    def recalculate(self, event: Event) -> Event:
        task_count = event.update[ParamName.TASK_COUNT].value
        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        control_task_count = event.update[ParamName.CONTROL_TASK_COUNT].value
        control_task_param = next(p for p in self.params if p.name == ParamName.CONTROL_TASK_COUNT)
        value, _, _, disabled = control_task_param.get_values(task_count, quality_preset, control_task_count)
        control_task_count_update = Update(value, disabled)
        new_update = copy(event.update)
        new_update[ParamName.CONTROL_TASK_COUNT] = control_task_count_update
        new_event = Event(sender=event.sender, update=new_update)
        return new_event


@dataclass
class ParamGroupCheckedTasksRecalculate(ParamGroup):
    def recalculate(self, event: Event) -> Event:
        task_count = event.update[ParamName.TASK_COUNT].value
        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        check_sample = event.update[ParamName.CHECK_SAMPLE].value
        max_tasks_to_check = event.update[ParamName.MAX_TASKS_TO_CHECK].value
        max_tasks_to_check_param = next(p for p in self.params if p.name == ParamName.MAX_TASKS_TO_CHECK)
        value, _, _, disabled = max_tasks_to_check_param.get_values(
            quality_preset, task_count, check_sample, max_tasks_to_check
        )
        max_tasks_to_check_update = Update(value, disabled)
        new_update = copy(event.update)
        new_update[ParamName.MAX_TASKS_TO_CHECK] = max_tasks_to_check_update
        new_event = Event(sender=event.sender, update=new_update)
        return new_event


@dataclass
class Help(Param):
    name: ParamName = ParamName.HELP
    description: str = ''
    value: str = '<b>Help for interface elements will be here</b>'
    available_range: List[OptionType] = field(default_factory=list)
    recommended_range: List[OptionType] = field(default_factory=list)
    help_msg: str = ''

    def process(self, event: Event) -> List[Event]:
        assert event.sender == self.name.value
        self.value = event.update[self.name].value
        return []


@dataclass
class Stat(Param):
    value: float = 0.0
    available_range: List[OptionType] = field(default_factory=list)
    recommended_range: List[OptionType] = field(default_factory=list)
    help_msg: str = ''
    reverse: bool = False


@dataclass
class StatWithExtraText(Stat):
    value: Tuple[float, str] = 0.0, ''


@dataclass
class Characteristics(Param):
    value: List[str] = field(default_factory=list)
    name: ParamName = ParamName.CHARACTERISTICS
    description: Optional[str] = None
    available_range: List[OptionType] = field(default_factory=list)
    recommended_range: List[OptionType] = field(default_factory=list)
    help_msg: str = ''


@dataclass(init=False)
class TaskDuration(Param):
    name: ParamName = ParamName.TASK_DURATION_HINT
    description: str = 'Task duration hint, sec'
    value: timedelta = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'This is task duration hint you provided (in seconds)'
    disabled: bool = True

    def __init__(self, task_duration_hint: timedelta):
        self.value = task_duration_hint
        self.available_range = [1, max(3 * 60, int(task_duration_hint.total_seconds()))]
        self.recommended_range = [5, 60]


@dataclass(init=False)
class SpeedControlList(Param):
    name: ParamName = ParamName.SPEED_CONTROL
    description: str = 'Speed control'
    value: List[BlockTimePicker] = field(init=False)
    available_range: Optional[list] = None
    recommended_range: Optional[list] = None
    help_msg: str = (
        'If workers complete your assignments suspiciously fast, you can ban them for some time and reject'
        ' these assignments. '
        'Threshold specifies percentage of time spent on task out of <i>task duration hint</i> total.'
    )
    disabled: bool = True
    list_size: int = 5

    min_value: int = pd.Timedelta('10m').total_seconds()
    max_value: int = pd.Timedelta('30d').total_seconds()

    @staticmethod
    def get_values(
        quality_preset: str, items: Optional[List[BlockTimePicker]] = None
    ) -> Tuple[List[BlockTimePicker], bool]:
        if quality_preset == QualityPresetType.NO_CONTROL.value:
            return [], True
        if quality_preset == QualityPresetType.CUSTOM.value:
            return (items or []), False
        speed_poor, speed_fraud = {
            QualityPresetType.MILD.value: (0.05, 0.01),
            QualityPresetType.MODERATE.value: (0.3, 0.1),
            QualityPresetType.STRICT.value: (0.5, 0.3),
        }[quality_preset]
        return [
            BlockTimePicker(speed_fraud, '1d', True),
            BlockTimePicker(speed_poor, '2h', False),
        ], True

    def __init__(self, quality_preset: str, list_size: int = 5):
        self.value, self.disabled = self.get_values(quality_preset)
        self.list_size = list_size

    def check_update(self, event: SpeedControlEvent) -> bool:
        for name, update in event.update.items():
            if (
                name in [SpeedControlChangeType.ADD, SpeedControlChangeType.REMOVE]
                or getattr(self.value[update.index], name.value) != update.value
            ):
                return True
        return False

    def process(self, event: Event) -> List[Event]:
        if event.sender == ParamName.GENERAL_QUALITY_GROUP.value:
            self.value, self.disabled = self.get_values(event.update[ParamName.QUALITY_PRESET].value, self.value)
            return []

        assert event.sender == self.name.ui
        assert len(event.update) == 1

        [(name, update)] = event.update.items()

        if name == SpeedControlChangeType.ADD:
            self.value.append(copy(default_block))
            return []

        index = update.index

        if name == SpeedControlChangeType.REMOVE:
            self.value.pop(index)
        else:
            setattr(self.value[index], name.value, update.value)
        return []


@dataclass(init=False)
class Lang(Param):
    name: ParamName = ParamName.LANGUAGE
    description: str = 'Language'
    value: str = field(init=False)
    available_range: List[str] = field(init=False)
    recommended_range: List[str] = field(init=False)
    help_msg: str = 'This is source language you provided for this parameters optimization'
    disabled: bool = True

    def __init__(self, task_spec: client.PreparedTaskSpec):
        self.value = task_spec.lang
        self.available_range = [self.value]
        self.recommended_range = ['EN', 'RU']


@dataclass(init=False)
class TaskCount(Param):
    name: ParamName = ParamName.TASK_COUNT
    description: str = 'Task count'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'Number of real tasks per assignment'

    def __init__(self, task_duration_hint: timedelta):
        max_task_count_calc = ceil(5 * 60 / task_duration_hint.total_seconds())
        min_task_count_calc = ceil(30 / task_duration_hint.total_seconds())

        min_task_count = max(min_tasks_in_assignment, min_task_count_calc)
        max_task_count = min(max_tasks_in_assignment // 2, max_task_count_calc)

        assert min_task_count <= max_task_count

        self.value = (min_task_count + max_task_count) // 2
        self.available_range = [min_tasks_in_assignment, max_tasks_in_assignment]
        self.recommended_range = [min_task_count, max_task_count]


def get_price_step(assignment_price: float) -> float:
    if assignment_price > threshold_cost:
        return 0.01
    return 0.001


def get_min_price(assignment_price: float) -> float:
    if assignment_price > threshold_cost:
        return 0.01
    return min_assignment_cost


@dataclass(init=False)
class AssignmentPrice(Param):
    name: ParamName = ParamName.ASSIGNMENT_PRICE
    description: str = 'Assignment price, $'
    value: float = field(init=False)
    available_range: List[float] = field(init=False)
    recommended_range: List[float] = field(init=False)
    help_msg: str = 'Base price of assignment that workers will receive, doesn\'t include Toloka commission'

    @staticmethod
    def get_values(
        task_duration_hint: timedelta,
        task_count: int,
        control_task_count: int,
    ):
        price_per_hour_range = [0.1, 2.5]
        default_price_per_hour_recommended_range = [0.3, 0.9]

        min_price = min_assignment_cost
        _, max_price, _ = _get_assignment_params(
            max_tasks_in_assignment + get_max_control_tasks_count(max_tasks_in_assignment),
            task_duration_hint,
            price_per_hour_range[-1],
        )

        _, min_rec_price, _ = _get_assignment_params(
            task_count + control_task_count,
            task_duration_hint,
            default_price_per_hour_recommended_range[0],
        )
        _, max_rec_price, _ = _get_assignment_params(
            task_count + control_task_count,
            task_duration_hint,
            default_price_per_hour_recommended_range[-1],
        )
        value = round_price((min_rec_price + max_rec_price) / 2)

        # we need to adjust min bound for parameter, so that current value can be represented as min + k * step
        # for prices above 1$ step is 0.01, so new min_assignment_cost=0.005 will break the logic
        min_price = max(min_price, get_min_price(value))
        return value, [min_price, max_price], [min_rec_price, max_rec_price]

    def __init__(self, task_duration_hint, task_count: int, control_task_count: int):
        self.value, self.available_range, self.recommended_range = self.get_values(
            task_duration_hint, task_count, control_task_count
        )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            if self.name in event.update:
                value, self.disabled = event.update[self.name].value, event.update[self.name].disabled
                self.available_range[0] = get_min_price(value)
                # we need to calculate step based on some value that is greater that threshold_cost
                step = get_price_step(threshold_cost + 1)
                if threshold_cost < value < threshold_cost + step:
                    # edge effect due to only 1 step value allowed for widget
                    # we catch "incorrect" increase here and automatically fix it
                    value = threshold_cost + step

                self.value = round_price(value)
            return []
        task_count = event.update[ParamName.TASK_COUNT].value
        control_task_count = event.update[ParamName.CONTROL_TASK_COUNT].value
        task_duration_hint = event.update[ParamName.TASK_DURATION_HINT].value

        self.value, self.available_range, self.recommended_range = self.get_values(
            task_duration_hint, task_count, control_task_count
        )
        return []


class AnnotationAssignmentPrice(AssignmentPrice):
    @staticmethod
    def get_values(
        task_duration_hint: timedelta,
        task_count: int,
    ):
        price_per_hour_range = [0.1, 2.5]
        default_price_per_hour_recommended_range = [0.3, 0.9]

        min_price = min_assignment_cost
        _, max_price, _ = _get_assignment_params(
            max_tasks_in_assignment,
            task_duration_hint,
            price_per_hour_range[-1],
        )

        _, min_rec_price, _ = _get_assignment_params(
            task_count,
            task_duration_hint,
            default_price_per_hour_recommended_range[0],
        )
        _, max_rec_price, _ = _get_assignment_params(
            task_count,
            task_duration_hint,
            default_price_per_hour_recommended_range[-1],
        )
        value = round_price((min_rec_price + max_rec_price) / 2)
        min_price = max(min_price, get_min_price(value))
        return value, [min_price, max_price], [min_rec_price, max_rec_price]

    def __init__(self, task_duration_hint, task_count: int):
        # todo: maybe we need to take # of checks per page into account
        self.value, self.available_range, self.recommended_range = self.get_values(task_duration_hint, task_count)

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)
        task_count = event.update[ParamName.TASK_COUNT].value
        task_duration_hint = event.update[ParamName.TASK_DURATION_HINT].value

        self.value, self.available_range, self.recommended_range = self.get_values(
            task_duration_hint,
            task_count,
        )
        return []


@dataclass
class QualityPreset(Param):
    name: ParamName = ParamName.QUALITY_PRESET
    description: str = 'Quality control preset'
    value: str = QualityPresetType.MODERATE.value
    available_range: List[str] = field(default_factory=lambda: [e.value for e in QualityPresetType])
    recommended_range: List[str] = field(
        default_factory=lambda: [QualityPresetType.MODERATE.value, QualityPresetType.STRICT.value]
    )
    help_msg: str = 'General quality control presets, that will manage all the options for you'


@dataclass
class AnnotationQualityPreset(QualityPreset):
    available_range: List[str] = field(
        default_factory=lambda: [e.value for e in QualityPresetType if e != QualityPresetType.NO_CONTROL]
    )


def get_max_control_tasks_count(task_count: int) -> int:
    return task_count


def get_checked_tasks_count(task_count: int, check_sample: bool, max_tasks_to_check: int) -> int:
    if not check_sample:
        return task_count
    assert max_tasks_to_check <= task_count
    return max_tasks_to_check


@dataclass(init=False)
class ControlTaskCount(Param):
    name: ParamName = ParamName.CONTROL_TASK_COUNT
    description: str = 'Control task count'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'Number of control tasks per assignment'

    @staticmethod
    def get_values(
        task_count: int,
        quality_preset: str,
        control_task_count: Optional[int] = None,
    ):
        control_task_count_rec_range = [1, ceil(0.25 * task_count)]
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            control_task_ratio = {
                QualityPresetType.NO_CONTROL.value: 0.0,
                QualityPresetType.MILD.value: 0.1,
                QualityPresetType.MODERATE.value: 0.2,
                QualityPresetType.STRICT.value: 0.4,
            }[quality_preset]

            control_task_count = ceil(control_task_ratio * task_count)

            # this is separate logic from get_max_control_tasks_count:
            # in this case, some preset is chosen, and some control_task_count is created automatically,
            # but we want to set an upper-bound to number of control tasks per page
            control_task_count_range = [0, max(control_task_count, min(10, task_count))]

        else:
            if control_task_count is None:
                control_task_count = (control_task_count_rec_range[0] + control_task_count_rec_range[-1]) // 2
            control_task_count_range = [0, get_max_control_tasks_count(task_count)]

        control_task_count_rec_range[-1] = min(control_task_count_rec_range[-1], control_task_count_range[-1])

        control_task_count = max(control_task_count, control_task_count_range[0])
        control_task_count = min(control_task_count, control_task_count_range[-1])

        return control_task_count, control_task_count_range, control_task_count_rec_range, disabled

    def __init__(self, task_count: int, quality_preset: str):
        self.value, self.available_range, self.recommended_range, self.disabled = self.get_values(
            task_count,
            quality_preset,
        )

    def process(self, event: Event) -> List[Event]:
        task_count = event.update[ParamName.TASK_COUNT].value
        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        control_task_count = event.update[ParamName.CONTROL_TASK_COUNT].value
        self.value, self.available_range, self.recommended_range, self.disabled = self.get_values(
            task_count, quality_preset, control_task_count
        )
        return []


@dataclass(init=False)
class ControlTasksForAccept(Param):
    name: ParamName = ParamName.CONTROL_TASKS_FOR_ACCEPT
    description: str = 'Accept, if correct control tasks &ge;'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = (
        'If worker completed correctly this many control tasks of more in the assignment, '
        'this assignment will be accepted; otherwise - rejected'
    )

    num_classes: int = field(init=False)

    @staticmethod
    def get_values(
        num_classes: int,
        quality_preset: str,
        control_task_count: int,
        control_tasks_for_accept: Optional[int] = None,
    ) -> Tuple[int, bool, List[int], List[int]]:
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            robustness_ = {
                QualityPresetType.NO_CONTROL.value: 0.0,
                QualityPresetType.MILD.value: 0.6,
                QualityPresetType.MODERATE.value: 0.75,
                QualityPresetType.STRICT.value: 0.9,
            }[quality_preset]

            for control_tasks_for_accept in range(control_task_count + 1):
                robustness = 1.0 - _get_random_chance(control_task_count, control_tasks_for_accept, num_classes)
                if robustness >= robustness_:
                    break
        control_tasks_for_accept = min(control_task_count, control_tasks_for_accept)
        control_task_for_accept_range = [0, control_task_count]
        control_task_for_accept_rec_range = [ceil(0.2 * control_task_count), ceil(0.6 * control_task_count)]
        return control_tasks_for_accept, disabled, control_task_for_accept_range, control_task_for_accept_rec_range

    def __init__(self, num_classes: int, quality_preset: str, control_task_count: int):
        self.num_classes = num_classes
        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            num_classes,
            quality_preset,
            control_task_count,
        )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        control_task_count = event.update[ParamName.CONTROL_TASK_COUNT].value

        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            self.num_classes,
            quality_preset,
            control_task_count,
            self.value,
        )
        return []


class CheckedTasksForAccept(ControlTasksForAccept):
    description: str = 'Accept, if correctly done checked tasks &ge;'
    num_classes: Optional[int] = None

    @staticmethod
    def get_values(
        num_classes: Optional[int],  # None, not used here, left for compatibility
        quality_preset: str,
        control_task_count: int,
        control_tasks_for_accept: Optional[int] = None,
    ) -> Tuple[int, bool, List[int], List[int]]:
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            ratio = {
                QualityPresetType.NO_CONTROL.value: 0.0,
                QualityPresetType.MILD.value: 0.6,
                QualityPresetType.MODERATE.value: 0.75,
                QualityPresetType.STRICT.value: 0.9,
            }[quality_preset]
            control_tasks_for_accept = ceil(ratio * control_task_count)

        control_tasks_for_accept = min(control_task_count, control_tasks_for_accept)
        control_task_for_accept_range = [0, control_task_count]
        control_task_for_accept_rec_range = [ceil(0.2 * control_task_count), ceil(0.6 * control_task_count)]
        return control_tasks_for_accept, disabled, control_task_for_accept_range, control_task_for_accept_rec_range

    def __init__(self, quality_preset: str, control_task_count: int):
        super().__init__(None, quality_preset, control_task_count)

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        task_count = event.update[ParamName.TASK_COUNT].value
        max_tasks_to_check = event.update[ParamName.MAX_TASKS_TO_CHECK].value
        check_sample = event.update[ParamName.CHECK_SAMPLE].value
        control_task_count = get_checked_tasks_count(task_count, check_sample, max_tasks_to_check)

        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            None,  # todo for compatibility reasons
            quality_preset,
            control_task_count,
            self.value,
        )
        return []


@dataclass(init=False)
class ControlTasksForBlock(Param):
    name: ParamName = ParamName.CONTROL_TASKS_FOR_BLOCK
    description: str = 'Block, if correct control tasks &le;'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = (
        'If worker completed correctly this many control tasks or less in the assignment, '
        'they would be temporarily blocked due to bad performance'
    )

    @staticmethod
    def get_values(
        quality_preset: str,
        control_task_count: int,
        control_tasks_for_block: Optional[int] = None,
    ):
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            block_ratio = {
                QualityPresetType.NO_CONTROL.value: 0.0,
                QualityPresetType.MILD.value: 0.1,
                QualityPresetType.MODERATE.value: 0.3,
                QualityPresetType.STRICT.value: 0.5,
            }[quality_preset]

            control_tasks_for_block = ceil(block_ratio * control_task_count)
        control_tasks_for_block = min(control_task_count, control_tasks_for_block)
        control_task_for_block_range = [0, control_task_count]
        control_task_for_block_rec_range = [0, ceil(0.3 * control_task_count)]

        return control_tasks_for_block, disabled, control_task_for_block_range, control_task_for_block_rec_range

    def __init__(self, quality_preset: str, control_task_count: int):
        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset,
            control_task_count,
        )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        control_task_count = event.update[ParamName.CONTROL_TASK_COUNT].value

        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset,
            control_task_count,
            self.value,
        )
        return []


class CheckedTasksForBlock(ControlTasksForBlock):
    description: str = 'Block, if correctly done checked tasks &le;'

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        task_count = event.update[ParamName.TASK_COUNT].value
        max_tasks_to_check = event.update[ParamName.MAX_TASKS_TO_CHECK].value
        check_sample = event.update[ParamName.CHECK_SAMPLE].value
        control_task_count = get_checked_tasks_count(task_count, check_sample, max_tasks_to_check)

        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset,
            control_task_count,
            self.value,
        )
        return []


@dataclass(init=False)
class Training(Param):
    name: ParamName = ParamName.TRAINING
    description: str = 'Training'
    value: bool = field(init=False)
    available_range: List[bool] = field(init=False)
    recommended_range: List[bool] = field(init=False)
    help_msg: str = 'Only workers that passed your training will be able to see your task'

    def __init__(self, task_spec: client.PreparedTaskSpec, toloka_client: toloka.TolokaClient):
        has_training = False
        try:
            result = client.find_project(task_spec, toloka_client)
            if result is not None:
                prj, _ = result
                training = experts.find_training(prj.id, toloka_client)
                has_training = training is not None
        except:
            pass
        if not has_training:
            self.help_msg = self.help_msg + '. You don\'t have any available trainings for this task'
        self.value = has_training
        self.disabled = not has_training
        self.available_range = [False, True]
        self.recommended_range = [True]

    def process(self, event: Event) -> List[Event]:
        return super().process(event) + [Event(sender=self.name.value, update={self.name: event.update[self.name]})]


@dataclass(init=False)
class CheckSample(Param):
    name: ParamName = ParamName.CHECK_SAMPLE
    description: str = 'Check sample'
    value: bool = field(init=False)
    available_range: List[bool] = field(init=False)
    recommended_range: List[bool] = field(init=False)
    help_msg: str = 'We can check only a subsample of submitted annotations, which cuts down cost of labeling'

    @staticmethod
    def get_values(quality_preset: str, check_sample: Optional[bool] = None) -> Tuple[bool, bool]:
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            value = False
        else:
            value = check_sample
        return value, disabled

    def __init__(self, quality_preset: str):
        self.value, self.disabled = self.get_values(quality_preset)
        self.available_range = [False, True]
        self.recommended_range = [False]

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        check_sample = event.update[self.name].value
        self.value, self.disabled = self.get_values(quality_preset, check_sample)
        return []


@dataclass(init=False)
class MaxTasksToCheck(Param):
    name: ParamName = ParamName.MAX_TASKS_TO_CHECK
    description: str = 'Max tasks to check'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'Maximum number of annotation tasks that would be checked from each assignment'

    @staticmethod
    def get_values(
        quality_preset: str, task_count: int, check_sample: bool, max_tasks_to_check: Optional[int] = None
    ) -> Tuple[int, bool, List[int], List[int]]:
        disabled = quality_preset != QualityPresetType.CUSTOM.value or not check_sample
        available_range = [1, task_count]
        recommended_range = [task_count]

        if disabled:
            value = task_count
        else:
            value = max_tasks_to_check
        return value, disabled, available_range, recommended_range

    def __init__(self, quality_preset: str, task_count: int, check_sample: bool):
        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset, task_count, check_sample
        )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)
        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        task_count = event.update[ParamName.TASK_COUNT].value
        check_sample = event.update[ParamName.CHECK_SAMPLE].value
        max_tasks_to_check = event.update[self.name].value
        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset,
            task_count,
            check_sample,
            max_tasks_to_check,
        )
        return []


@dataclass(init=False)
class AccuracyThreshold(Param):
    name: ParamName = ParamName.ACCURACY_THRESHOLD
    description: str = 'Accuracy threshold'
    value: int = field(init=False)
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'Threshold number of correctly done tasks in assignment to finalize it'

    @staticmethod
    def get_values(
        quality_preset: str,
        task_count: int,
        check_sample: bool,
        max_tasks_to_check: int,
        accuracy_threshold: Optional[int] = None,
    ) -> Tuple[int, bool, List[int], List[int]]:
        disabled = quality_preset != QualityPresetType.CUSTOM.value or not check_sample
        max_task_count = get_checked_tasks_count(task_count, check_sample, max_tasks_to_check)
        if disabled:
            value = max_task_count
        else:
            value = accuracy_threshold
        available_range = [1, max_task_count]
        recommended_range = [ceil(0.75 * max_task_count), ceil(0.95 * max_task_count)]

        return value, disabled, available_range, recommended_range

    def __init__(self, quality_preset: str, task_count: int, check_sample: bool, max_tasks_to_check: int):
        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset, task_count, check_sample, max_tasks_to_check
        )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)
        quality_preset = event.update[ParamName.QUALITY_PRESET].value
        task_count = event.update[ParamName.TASK_COUNT].value
        check_sample = event.update[ParamName.CHECK_SAMPLE].value
        max_tasks_to_check = event.update[ParamName.MAX_TASKS_TO_CHECK].value

        self.value, self.disabled, self.available_range, self.recommended_range = self.get_values(
            quality_preset, task_count, check_sample, max_tasks_to_check, self.value
        )
        return []


@dataclass(init=False)
class TrainingValue(Param):
    name: ParamName = ParamName.TRAINING_SCORE
    description: str = 'Training score'
    value: int = 80
    available_range: List[int] = field(init=False)
    recommended_range: List[int] = field(init=False)
    help_msg: str = 'Percentage of training that should be completed correctly to qualify as "passed"'

    def __init__(self, training: bool):
        self.disabled = not training
        self.available_range = [0, 100]
        self.recommended_range = [75, 95]

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        self.disabled = not event.update[ParamName.TRAINING].value
        return []


@dataclass(init=False)
class AgeRestriction(Param):
    name: ParamName = ParamName.AGE_RESTRICTION
    description: str = '18+'
    value: bool = True
    available_range: List[bool] = field(init=False)
    recommended_range: List[bool] = field(init=False)
    help_msg: str = (
        'If there may be something inappropriate in your tasks, you should only show them to mature audience'
    )

    def __init__(self):
        self.available_range = [False, True]
        self.recommended_range = [True]


@dataclass(init=False)
class VerifiedLang(Param):
    name: ParamName = ParamName.VERIFIED_LANG
    description: str = 'Verified language'
    value: bool = field(init=False)
    available_range: List[bool] = field(init=False)
    recommended_range: List[bool] = field(init=False)
    help_msg: str = (
        'Only workers that completed extra test and verified their knowledge of your chosen language '
        'will be able to see your task'
    )

    def __init__(self, task_spec: client.PreparedTaskSpec):
        lang = task_spec.lang
        self.available_range = [False, True]
        self.recommended_range = [True]

        if lang in Languages.verified_languages_to_skills:
            self.value = True
        else:
            self.value = False
            self.disabled = True
            self.help_msg = self.help_msg + '. This option is not available for your chosen language'


@dataclass(init=False)
class Overlap(Param):
    name: ParamName = ParamName.OVERLAP
    description: str = 'Overlap'
    value: str = field(init=False)
    available_range: List[str] = field(init=False)
    recommended_range: List[str] = field(init=False)
    help_msg: str = (
        'Number of accepted solutions that will be collected for each individual task. '
        'Can be static - e.g. "3" or dynamic - e.g. "3-5". Custom overlap values are supported.'
    )

    @staticmethod
    def get_values(
        quality_preset: str,
        overlap: Optional[str] = None,
    ):
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            overlap = {
                QualityPresetType.NO_CONTROL.value: '1',
                QualityPresetType.MILD.value: '2-3',
                QualityPresetType.MODERATE.value: '3',
                QualityPresetType.STRICT.value: '3-5',
            }[quality_preset]

        else:
            if overlap is None:
                overlap = '3'

        return overlap, disabled

    def __init__(self, quality_preset: str):
        self.available_range = ['1', '2-3', '3', '3-5', '5']
        self.recommended_range = ['2-3', '3']

        self.value, self.disabled = self.get_values(quality_preset)

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            update = event.update[self.name]
            try:
                get_overlap(update.value, 0.5)
                return super().process(event) + [Event(sender=self.name.value, update={self.name: update})]
            except:
                return []

        quality_preset = event.update[ParamName.QUALITY_PRESET].value

        self.value, self.disabled = self.get_values(
            quality_preset,
            self.value,
        )
        return [Event(sender=self.name.value, update={self.name: Update(self.value, self.disabled)})]


class MaxAttempts(Overlap):
    description: str = 'Max attempts'

    help_msg: str = (
        'Maximum number of annotation attempts that will be collected for each individual task. '
        'Custom overlap values are supported.'
    )

    @staticmethod
    def get_values(
        quality_preset: str,
        max_attempts: Optional[str] = None,
    ):
        disabled = quality_preset != QualityPresetType.CUSTOM.value
        if disabled:
            max_attempts = {
                QualityPresetType.MILD.value: '2',
                QualityPresetType.MODERATE.value: '3',
                QualityPresetType.STRICT.value: '5',
            }[quality_preset]

        else:
            if max_attempts is None:
                max_attempts = '3'

        return max_attempts, disabled

    def __init__(self, quality_preset: str):
        self.available_range = ['2', '3', '5', '7']
        self.recommended_range = ['2', '3']

        self.value, self.disabled = self.get_values(quality_preset)

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            update = event.update[self.name]
            try:
                assert int(update.value) > 0
                return super().process(event) + [Event(sender=self.name.value, update={self.name: update})]
            except:
                return []

        quality_preset = event.update[ParamName.QUALITY_PRESET].value

        self.value, self.disabled = self.get_values(
            quality_preset,
            self.value,
        )
        return [Event(sender=self.name.value, update={self.name: Update(self.value, self.disabled)})]


@dataclass(init=False)
class OverlapConfidence(Param):
    name: ParamName = ParamName.OVERLAP_CONFIDENCE
    description: str = 'Dynamic overlap confidence'
    value: float = 0.85
    available_range: List[float] = field(init=False)
    recommended_range: List[float] = field(init=False)
    help_msg: str = (
        'Applicable for dynamic overlap only. Threshold confidence of solution to qualify as "good", '
        'which can happen before number of attempts reaches maximum'
    )

    def __init__(self, quality_preset: str, overlap: str):
        self.disabled = quality_preset != QualityPresetType.CUSTOM.value or overlap_is_static(overlap)
        self.available_range = [0.0, 1.0]
        self.recommended_range = [0.75, 0.90]

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        overlap_update = event.update[ParamName.OVERLAP]
        disabled = overlap_update.disabled or overlap_is_static(overlap_update.value)

        if disabled:
            self.value, self.disabled = 0.85, True
        else:
            self.disabled = False
        return []


class Confidence(OverlapConfidence):
    description: str = 'Annotation confidence'
    help_msg: str = (
        'Threshold confidence of annotation to qualify as "good", '
        'which can happen before number of attempts reaches maximum'
    )

    def process(self, event: Event) -> List[Event]:
        if isinstance(event, UIEvent):
            return super().process(event)

        overlap_update = event.update[ParamName.OVERLAP]
        disabled = overlap_update.disabled

        if disabled:
            self.value, self.disabled = 0.85, True
        else:
            self.disabled = False
        return []


@dataclass(init=False)
class AggregationAlgorithm(Param):
    name: ParamName = ParamName.AGGREGATION_ALGORITHM
    description: str = 'Aggregation'
    value: str = 'DS'
    available_range: List[str] = field(init=False)
    recommended_range: List[str] = field(init=False)
    help_msg: str = (
        'Aggregation algorithm that will be applied to solutions. '
        'Options are: Majority Vote, '
        '<a href="http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.469.1377&rep=rep1&type=pdf"> Dawid-Skene</a>'
        ' and Max Likelihood'
    )

    def __init__(self):
        self.available_range = ['MV', 'ML', 'DS']
        self.recommended_range = ['ML', 'DS']



import os

from setuptools import setup, find_packages

PREFIX = 'dataforge'

setup_py_dir = os.path.dirname(__file__)
version_module_path = os.path.join(setup_py_dir, 'src', '__version__.py')

about = {}

with open(version_module_path) as f:
    exec(f.read(), about)

with open('README.md') as f:
    readme = f.read()

setup(
    name=about['__title__'],
    package_dir={PREFIX: 'src'},
    packages=['dataforge', *(f'{PREFIX}.{package}' for package in find_packages('src'))],
    version=about['__version__'],
    description='DataForge',
    long_description=readme,
    long_description_content_type='text/markdown',
    license=about['__license__'],
    author='Oleg Gulyaev',
    author_email='o-gulyaev@yandex-team.ru',
    python_requires='>=3.8.0',
    install_requires=[
        'toloka-kit == 0.1.26',
        'crowd-kit == 1.0.0',
        'python-dateutil >= 2.8.2',
        'beautifulsoup4 >= 4.8.2',
        'matplotlib >= 3.5.0',
        'pylzy == 0.0.49',
        'ipywidgets >= 7.7.0',
    ],
    include_package_data=True,
)



from dataclasses import dataclass
from typing import Optional

from ..classification import AggregationAlgorithm
from ..classification_loop import Overlap, DynamicOverlap
from ..control import Control
from ..evaluation import AssignmentCheckSample
from ..duration import TaskDurationFunction


@dataclass
class Evaluation:
    aggregation_algorithm: AggregationAlgorithm
    assignment_check_sample: Optional[AssignmentCheckSample]
    overlap: Overlap

    def __post_init__(self):
        check_sample = self.assignment_check_sample
        if check_sample is not None and check_sample.max_tasks_to_check is not None:
            assert check_sample.max_tasks_to_check > 0


@dataclass
class Params:
    assignment_check_sample: Optional[AssignmentCheckSample]
    control: Control
    overlap: DynamicOverlap
    task_duration_function: TaskDurationFunction



from .. import base

annotation_check_cut = base.LocalizedString(
    {
        'EN': """
<div>
    <p>
        You will see solutions to tasks completed by the instruction below. Evaluate solutions according to this instruction.
    </p>
</div>
<h1>Instructions according to which solutions were given</h1>
""",
        'RU': """
<div>
    <p>
        Вы увидите решения заданий, выполненные по инструкции ниже. Проверьте решения на соответствие инструкции.
    </p>
</div>
<h1>Инструкция, по которой выполнены решения</h1>
""",
        'TR': """
<div>
    <p>
        Aşağıdaki talimatlara göre tamamlanan iş çözümlerini göreceksiniz. Kontrol çözümleri uyum için talimatlar.
    </p>
</div>
<h1>Çözümlerin yürütüldüğü talimat</h1>
""",
        'KK': """
<div>
    <p>
        Төмендегі нұсқауларға сәйкес орындалған тапсырмалардың шешімдерін көресіз. Шешімдердің нұсқаулыққа сәйкестігін тексеріңіз.
    </p>
</div>
<h1>Шешімдер орындалған Нұсқаулық</h1>
""",
    }
)



from collections import Counter, defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import List, Tuple, Iterable, Optional, Type

import toloka.client.project.template_builder as tb
import toloka.client as toloka

from ..base import (
    TaskFunction,
    ClassificationFunction,
    AnnotationFunction,
    SbSFunction,
    BinaryEvaluation,
    EvaluationMeta,
    Label,
    Object,
    ObjectMeta,
    LocalizedString,
    TextFormat,
    Title,
    ConditionContext,
    ClassMeta,
    ImageAnnotation,
    ImageAnnotationMeta,
)
from ..mapping import TaskMapping, ObjectMapping
from ..objects import Audio, Image, Video, Text, TextMeta
from .viewers import object_to_info, media_notifications


class Scenario(Enum):
    DEFAULT = 'default'
    EXPERT_LABELING_OF_TASKS = 'expert-task'
    EXPERT_LABELING_OF_SOLVED_TASKS = 'expert-solution'

    @property
    def project_name_suffix(self) -> LocalizedString:
        return LocalizedString(
            {
                self.EXPERT_LABELING_OF_TASKS: {'EN': 'task labeling', 'RU': 'разметка заданий'},
                self.EXPERT_LABELING_OF_SOLVED_TASKS: {'EN': 'solution labeling', 'RU': 'разметка решений'},
            }[self]
        )


@dataclass
class Builder:
    task_function: TaskFunction
    lang: str
    current_hotkey: int = 1
    task_width: float = 700.0

    condition_context: ConditionContext = ConditionContext(name_to_arg_data={})

    EXPERT_LABELING_METAS = (
        EvaluationMeta(
            BinaryEvaluation,
            '_ok',
            Title(
                text=LocalizedString(
                    {
                        'EN': 'Is the task (and the solution, if given) correct?',
                        'RU': 'Корректно ли задание (и решение, при наличии)?',
                    },
                ),
            ),
        ),
        ObjectMeta(
            Text,
            '_comment',
            Title(text=LocalizedString({'EN': 'comment', 'RU': 'комментарий'})),
            required=False,
        ),
    )

    def get_view_and_mapping(
        self,
        items: Tuple[ObjectMeta, ...],
        is_output: bool = False,
        suffix: str = '',
    ) -> Tuple[List[tb.view.BaseViewV1], List[ObjectMapping], List[tb.plugins.HotkeysPluginV1]]:
        object_views, object_mappings, hotkeys = [], [], []
        for obj_view, obj_mapping, obj_hotkeys in self.get_object_list_tb_views_and_mappings(
            items,
            is_output=is_output,
            suffix=suffix,
        ):
            object_views.append(obj_view)
            object_mappings.append(obj_mapping)
            hotkeys.extend(obj_hotkeys)
        return object_views, object_mappings, hotkeys

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.task_function.get_inputs()

    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        return self.task_function.get_outputs()

    def get_object_tb_view_and_mapping(
        self,
        obj_meta: ObjectMeta,
        is_output: bool,
        suffix: str = '',
    ) -> Tuple[tb.view.BaseViewV1, ObjectMapping, List[tb.plugins.HotkeysPluginV1]]:
        info = object_to_info[obj_meta.type]
        task_field = f'{obj_meta.name}{suffix}'
        obj_field = info.obj_field
        view_field = info.view_field

        self.condition_context.name_to_arg_data[obj_meta.name] = ConditionContext.ArgData(
            final_name=task_field,
            is_output=is_output,
        )

        obj_mapping = ObjectMapping(obj_meta=obj_meta, obj_task_fields=((obj_field, task_field),))

        hotkeys_enabled = True
        if isinstance(obj_meta, ClassMeta) and obj_meta.available_labels:
            # TODO (DATAFORGE-70): correct hotkeys for available labels
            hotkeys_enabled = False

        if is_output:
            data = tb.OutputData(path=task_field)
            validation = info.output_validation
            kwargs = {'data': data}
            viewer = info.output_viewer
            hotkeys = []
            if hotkeys_enabled:
                hotkeys = info.hotkeys(obj_meta.type, start_key=self.current_hotkey, output_path=task_field)
                self.current_hotkey += len(hotkeys)
        else:
            if issubclass(obj_meta.type, Label) or issubclass(obj_meta.type, ImageAnnotation):
                # radio group button cannot be used with InputData directly
                data = {'data': tb.InternalData(path=task_field, default=tb.InputData(path=task_field))}
            else:
                data = {view_field: tb.InputData(path=task_field)}
            validation = info.input_validation
            kwargs = {**data}
            viewer = info.input_viewer
            hotkeys = []

        if validation:
            # specifying 'data' is not always necessary, TB can determine data itself
            # but for some cases, i.e. compare layout in SbS, we have to specify data
            kwargs['validation'] = validation(data=data, lang=self.lang)

        if issubclass(obj_meta.type, Label):
            kwargs['lang'] = self.lang
            kwargs['obj_meta'] = obj_meta
            kwargs['condition_context'] = self.condition_context

        if isinstance(obj_meta, ImageAnnotationMeta):
            kwargs['shapes'] = {shape: True for shape in obj_meta.available_shapes}
            if obj_meta.labels is not None:
                kwargs['labels'] = [
                    tb.fields.ImageAnnotationFieldV1.Label(label=item.get_label()[self.lang], value=item.value)
                    for item in obj_meta.labels.possible_instances()
                ]

        if issubclass(obj_meta.type, Text):
            if is_output:
                title = None
                validation = None
                if obj_meta.title:
                    title = obj_meta.title.text
                if isinstance(obj_meta, TextMeta) and obj_meta.validation:
                    validation = obj_meta.validation

                kwargs['placeholder'] = title
                assert 'validation' not in kwargs
                kwargs['validation'] = validation
                kwargs['lang'] = self.lang

            else:
                text_format = TextFormat.PLAIN
                if isinstance(obj_meta, TextMeta):
                    text_format = obj_meta.format
                kwargs['text_format'] = text_format

        return viewer(**kwargs), obj_mapping, hotkeys

    def get_object_list_tb_views_and_mappings(
        self,
        obj_metas: Tuple[ObjectMeta, ...],
        is_output: bool,
        suffix: str = '',
    ) -> Iterable[Tuple[tb.view.BaseViewV1, ObjectMapping, List[tb.plugins.HotkeysPluginV1]]]:
        # TODO: handle task field collisions in case of multiple objects of the same type
        for obj_single_meta in obj_metas:
            yield self.get_object_tb_view_and_mapping(obj_single_meta, is_output=is_output, suffix=suffix)

    def gen_media_notifications(self, obj_metas: Tuple[ObjectMeta, ...]) -> Optional[List[tb.TextViewV1]]:
        need_media_notifications = any(meta.type in [Audio, Video, Image] for meta in obj_metas)
        if need_media_notifications:
            return [tb.TextViewV1(content=notification[self.lang]) for notification in media_notifications]
        return None

    def get_plugins(
        self,
        hotkeys: List[tb.plugins.HotkeysPluginV1],
    ) -> List[tb.plugins.BasePluginV1]:
        return hotkeys + [
            tb.TolokaPluginV1(
                'scroll', task_width=self.task_width, notifications=self.gen_media_notifications(self.get_inputs())
            )
        ]

    def combine_views(
        self,
        views: List[tb.view.BaseViewV1],
        hotkeys: List[tb.plugins.HotkeysPluginV1],
        has_expert_fields: bool = False,
    ) -> toloka.project.view_spec.ViewSpec:
        if has_expert_fields:
            views = (
                views[: -len(self.EXPERT_LABELING_METAS)]
                + [tb.MarkdownViewV1(content='---')]
                + views[-len(self.EXPERT_LABELING_METAS) :]
            )
        return toloka.project.TemplateBuilderViewSpec(
            view=tb.ListViewV1(items=views), plugins=self.get_plugins(hotkeys)
        )

    def combine_mappings(self, input_obj_mappings: List[ObjectMapping], output_obj_mappings: List[ObjectMapping]):
        return TaskMapping(input_mapping=tuple(input_obj_mappings), output_mapping=tuple(output_obj_mappings))

    def get_scenario(
        self, scenario: Scenario, for_preview: bool = False
    ) -> Tuple[toloka.project.view_spec.ViewSpec, TaskMapping]:
        expert_meta = self.EXPERT_LABELING_METAS
        if not self.task_function.has_names:
            suffixes = get_obj_type_indexes(tuple(meta.type for meta in self.get_all_arguments()))

            input_meta = tuple(
                meta.replace_name(object_to_info[meta.type].task_field + suffix)
                for meta, suffix in zip(self.get_inputs(), suffixes[: len(self.get_inputs())])
            )
            output_meta = tuple(
                meta.replace_name(object_to_info[meta.type].task_field + suffix)
                for meta, suffix in zip(self.get_outputs(), suffixes[len(self.get_inputs()) :])
            )
        else:
            input_meta, output_meta = self.get_inputs(), self.get_outputs()

        if scenario == Scenario.DEFAULT:
            input_types, output_types = input_meta, output_meta
        elif scenario == Scenario.EXPERT_LABELING_OF_TASKS:
            input_types, output_types = input_meta, output_meta + expert_meta
        elif scenario == Scenario.EXPERT_LABELING_OF_SOLVED_TASKS:
            input_types, output_types = input_meta + output_meta, expert_meta
        else:
            raise ValueError(f'unsupported scenario: {scenario}')
        if for_preview:
            input_types, output_types = input_types + output_types, ()
        input_view, input_obj_mappings, _ = self.get_view_and_mapping(input_types, is_output=False)
        output_view, output_obj_mappings, hotkeys = self.get_view_and_mapping(output_types, is_output=True)
        return self.combine_views(
            input_view + output_view, hotkeys, has_expert_fields=scenario != Scenario.DEFAULT
        ), self.combine_mappings(input_obj_mappings, output_obj_mappings)

    def get_default_scenario(self) -> Tuple[toloka.project.view_spec.ViewSpec, TaskMapping]:
        return self.get_scenario(Scenario.DEFAULT)

    def get_expert_labeling_of_tasks(self) -> Tuple[toloka.project.view_spec.ViewSpec, TaskMapping]:
        return self.get_scenario(Scenario.EXPERT_LABELING_OF_TASKS)

    def get_expert_labeling_of_solved_tasks(self) -> Tuple[toloka.project.view_spec.ViewSpec, TaskMapping]:
        return self.get_scenario(Scenario.EXPERT_LABELING_OF_SOLVED_TASKS)

    def get_all_arguments(self) -> Tuple[ObjectMeta, ...]:
        """May include other types that should be considered when giving out task_field names"""
        return self.task_function.get_all_arguments()


class ClassificationBuilder(Builder):
    task_function: ClassificationFunction


class SbSBuilder(Builder):
    task_function: SbSFunction

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.task_function.get_hints() + self.task_function.get_inputs()

    def combine_views(
        self,
        views: List[tb.view.BaseViewV1],
        hotkeys: List[tb.plugins.HotkeysPluginV1],
        has_expert_fields: bool = False,
    ) -> toloka.project.view_spec.ViewSpec:
        # we should use sbs template for visually large objects
        sbs_layout = any(input_meta.type in [Image, Video] for input_meta in self.get_inputs())
        self.task_width = 1200.0 if sbs_layout else 700.0

        hint_type_count = len(self.task_function.get_hints())
        input_type_count = len(self.task_function.get_inputs())
        non_output_type_count = hint_type_count + 2 * input_type_count

        hint_view = views[:hint_type_count]
        sbs_input_view = views[hint_type_count:non_output_type_count]
        choice_view, other_output_view = views[non_output_type_count], views[non_output_type_count + 1 :]

        views_a, views_b = sbs_input_view[:input_type_count], sbs_input_view[input_type_count:]

        def gen_sbs_objects_view(views: List[tb.view.BaseViewV1]) -> tb.view.BaseViewV1:
            if len(views) == 1:
                return views[0]
            return tb.ListViewV1(items=views)

        views_a, views_b = (gen_sbs_objects_view(views) for views in (views_a, views_b))

        if sbs_layout:
            sbs_view = tb.SideBySideLayoutV1(items=[views_a, views_b], controls=choice_view)
        else:
            common_controls = tb.ColumnsLayoutV1(
                min_width=10.0,
                vertical_align='top',
                items=[
                    tb.RadioGroupFieldV1(data=choice_view.data, disabled=choice_view.disabled, options=[option])
                    for option in choice_view.options
                ],
                validation=choice_view.validation,
            )
            sbs_view = tb.CompareLayoutV1(
                items=[tb.CompareLayoutItem(content=view) for view in (views_a, views_b)],
                wide_common_controls=True,
                common_controls=common_controls,
            )

        return super(SbSBuilder, self).combine_views(
            hint_view + [sbs_view] + other_output_view, hotkeys, has_expert_fields
        )

    def get_view_and_mapping(
        self,
        items: Tuple[ObjectMeta, ...],
        is_output: bool = False,
        suffix: str = '',
    ) -> Tuple[List[tb.view.BaseViewV1], List[ObjectMapping], List[tb.plugins.HotkeysPluginV1]]:
        if is_output:
            return super(SbSBuilder, self).get_view_and_mapping(items, is_output, suffix)

        hint_type_count = len(self.task_function.get_hints())
        input_type_count = len(self.task_function.get_inputs())

        hint_types, input_types = items[:hint_type_count], items[hint_type_count : hint_type_count + input_type_count]

        views_a, views_b, views_hint = [], [], []
        mappings_a, mappings_b, mappings_hint = [], [], []
        for views, mappings, suffix, types in (
            (views_hint, mappings_hint, '_hint', hint_types),
            (views_a, mappings_a, '_a', input_types),
            (views_b, mappings_b, '_b', input_types),
        ):
            for view, obj_mapping, _ in self.get_object_list_tb_views_and_mappings(
                types,
                is_output=False,
                suffix=suffix,
            ):
                views.append(view)
                mappings.append(obj_mapping)

        if hint_type_count + input_type_count < len(items):
            choice = items[-1]
            choice_object_view, choice_object_mapping, _ = self.get_object_tb_view_and_mapping(
                choice, is_output=is_output
            )
            choice_view = [choice_object_view]
            choice_mapping = [choice_object_mapping]
        else:
            choice_view, choice_mapping = [], []

        return (
            views_hint + views_a + views_b + choice_view,
            mappings_hint + mappings_a + mappings_b + choice_mapping,
            [],
        )


class ImageAnnotationMixin(Builder):
    task_function: AnnotationFunction

    def get_plugins(
        self,
        hotkeys: List[tb.plugins.HotkeysPluginV1],
    ) -> List[tb.plugins.BasePluginV1]:
        if not self.task_function.has_image_annotation():
            return super(ImageAnnotationMixin, self).get_plugins(hotkeys)

        return [tb.plugins.ImageAnnotationHotkeysPluginV1()] + [
            tb.TolokaPluginV1('pager', notifications=self.gen_media_notifications(self.get_inputs()))
        ]

    def combine_views(
        self,
        views: List[tb.view.BaseViewV1],
        hotkeys: List[tb.plugins.HotkeysPluginV1],
        has_expert_fields: bool = False,
    ) -> toloka.project.view_spec.ViewSpec:
        if not self.task_function.has_image_annotation():
            return super(ImageAnnotationMixin, self).combine_views(views, hotkeys, has_expert_fields)
        image_index = None
        for i, view in enumerate(views):
            if isinstance(view, tb.view.ImageViewV1):
                if image_index is None:
                    image_index = i
                else:
                    assert False, 'Found multiple input images for image segmentation task'
            elif isinstance(view, tb.fields.ImageAnnotationFieldV1):
                assert image_index is not None, 'No input image found for image segmentation task'
                view.image = views[image_index].url

        assert image_index is not None, 'No input image found for image segmentation task'
        return super(ImageAnnotationMixin, self).combine_views(
            views[:image_index] + views[image_index + 1 :], hotkeys, has_expert_fields
        )


class AnnotationBuilder(ImageAnnotationMixin):
    task_function: AnnotationFunction

    def get_scenario(
        self, scenario: Scenario, for_preview: bool = False
    ) -> Tuple[toloka.project.view_spec.ViewSpec, TaskMapping]:
        if scenario == Scenario.EXPERT_LABELING_OF_TASKS:
            evaluation_meta = self.task_function.get_evaluation_named_meta()
            self.EXPERT_LABELING_METAS = (evaluation_meta,) + self.EXPERT_LABELING_METAS
        return super(AnnotationBuilder, self).get_scenario(scenario, for_preview)


class AnnotationCheckBuilder(ImageAnnotationMixin):
    task_function: AnnotationFunction

    def get_inputs(self) -> Tuple[ObjectMeta, ...]:
        return self.task_function.get_inputs() + self.task_function.get_outputs()

    def get_outputs(self) -> Tuple[ObjectMeta, ...]:
        return (self.task_function.get_evaluation(),)


# there can be many objects of the same type in input/output of function, we need to distinguish them by indexes
def get_obj_type_indexes(types: Tuple[Type[Object], ...]) -> List[str]:
    # we should use task_fields here, because different subclasses of Class will get the same 'choice' path
    task_fields = [object_to_info[type].task_field for type in types]
    single_obj_task_fields = {t for t, cnt in Counter(task_fields).items() if cnt == 1}
    task_field_counts = defaultdict(int)
    indexes = []
    for t in task_fields:
        task_field_counts[t] += 1
        indexes.append('' if t in single_obj_task_fields else f'_{task_field_counts[t]}')
    return indexes



from datetime import timedelta
from typing import Callable

from . import mapping

TaskDurationFunction = Callable[[mapping.Objects], timedelta]


def get_const_task_duration_function(task_duration_hint: timedelta) -> TaskDurationFunction:
    def task_duration_function(_):
        return task_duration_hint

    return task_duration_function



from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from typing import List, Dict, Union, Optional


class QualityPresetType(Enum):
    NO_CONTROL = 'No control'
    MILD = 'Mild'
    MODERATE = 'Moderate'
    STRICT = 'Strict'
    CUSTOM = 'Custom'


OptionType = Union[str, int, float, bool, timedelta, QualityPresetType, List['OptionType']]


class ParamName(Enum):
    HELP = 'help'
    CHARACTERISTICS = 'characteristics'
    TASK_DURATION_HINT = 'task_duration_hint'
    SPEED_CONTROL = 'speed_control'
    LANGUAGE = 'lang'
    TASK_COUNT = 'task_count'
    ASSIGNMENT_PRICE = 'assignment_price'
    QUALITY_PRESET = 'quality_preset'
    CONTROL_TASK_COUNT = 'control_task_count'
    CONTROL_TASKS_FOR_ACCEPT = 'control_tasks_for_accept'
    CONTROL_TASKS_FOR_BLOCK = 'control_tasks_for_block'
    TRAINING = 'training'
    TRAINING_SCORE = 'training_score'
    AGE_RESTRICTION = 'age_restriction'
    VERIFIED_LANG = 'verified_lang'
    OVERLAP = 'overlap'
    OVERLAP_CONFIDENCE = 'overlap_confidence'
    AGGREGATION_ALGORITHM = 'aggregation_algorithm'
    QUALITY = 'quality'
    TIME = 'time'
    COST = 'cost'
    GENERAL_QUALITY_GROUP = 'general_quality_group'
    QUALITY_TASKS_GROUP = 'quality_tasks_group'
    QUALITY_ASSIGNMENT_GROUP = 'quality_assignment_group'
    CHECK_SAMPLE_GROUP = 'check_sample_group'
    STATISTICS = 'statistics'
    CHECK_SAMPLE = 'check_sample'
    MAX_TASKS_TO_CHECK = 'max_tasks_to_check'
    ACCURACY_THRESHOLD = 'accuracy_threshold'

    @property
    def ui(self) -> str:
        return f'ui-{self.value}'


@dataclass
class Update:
    value: OptionType
    disabled: bool = False


@dataclass
class Event:
    sender: str
    update: Dict[ParamName, Update]


@dataclass
class UIEvent(Event):
    ...


class SpeedControlChangeType(Enum):
    ADD = 'add'
    REMOVE = 'remove'
    RATIO = 'ratio'
    REJECT = 'reject'
    TIME = 'time_value'


@dataclass
class SpeedControlUpdate(Update):
    change_type: SpeedControlChangeType = SpeedControlChangeType.ADD
    index: Optional[int] = None


@dataclass
class SpeedControlEvent(UIEvent):
    update: Dict[SpeedControlChangeType, SpeedControlUpdate]


@dataclass
class Node:
    name: ParamName

    def process(self, event: Event) -> List[Event]:
        return []



from .display import *  # noqa
from .launch import *  # noqa
from .params import *  # noqa
from .relaunch import *  # noqa
from .sbs import *  # noqa
from .task import *  # noqa
from .training import *  # noqa
from .utils import *  # noqa



from datetime import timedelta
import time
from typing import List, Tuple, Optional, Union

import numpy as np

from .. import classification_loop, client, evaluation, pricing
from ..pricing import _get_random_chance

from .util import get_overlap


ITERATIONS_COUNT: int = 500
CLASSIFICATION_WORKERS_COUNT = 7
ANNOTATION_WORKERS_COUNT = CLASSIFICATION_WORKERS_COUNT - 1


def hmean(a: list) -> float:
    return len(a) / (1 / np.maximum(np.array(a), 1e-12)).sum()


def simulate(
    overlap: classification_loop.Overlap,
    worker_q: float,
    control_task_count: int,
    control_tasks_for_accept: int,
    correct_control_tasks_in_solved: np.array,
    time_budget: float = 3.0,
    collect_rejected_attempts: bool = False,
    accuracy_threshold: Optional[int] = None,
) -> Tuple[int, int, int, float, bool]:
    start = time.time()
    ok = True
    overlap_increase = overlap.min_overlap
    needed_confidence = 0.0 if isinstance(overlap, classification_loop.StaticOverlap) else overlap.confidence
    if (
        collect_rejected_attempts and accuracy_threshold is not None
    ):  # we can have annotation setup, but no check sample
        needed_confidence = min(needed_confidence, accuracy_threshold / control_task_count)
    assert isinstance(needed_confidence, float)  # may be dict {label: float}
    accepted_solutions_count, loop_iterations, attempts = 0, 0, 0
    correct_control_tasks_in_accepted = np.array([], dtype=int)

    while overlap_increase > 0:
        loop_iterations += 1
        current = time.time()
        budget = time_budget - (current - start)
        if budget < 0 or attempts + overlap_increase >= len(correct_control_tasks_in_solved):
            ok = False
            break
        correct_control_tasks_in_collected = correct_control_tasks_in_solved[attempts : attempts + overlap_increase]
        attempts += overlap_increase

        accepted_solutions_count += (correct_control_tasks_in_collected >= control_tasks_for_accept).sum()

        if not collect_rejected_attempts:
            correct_control_tasks_in_collected = correct_control_tasks_in_collected[
                correct_control_tasks_in_collected >= control_tasks_for_accept
            ]

        correct_control_tasks_in_accepted = np.concatenate(
            (
                correct_control_tasks_in_accepted,
                correct_control_tasks_in_collected,
            )
        )

        collected_solutions_count = attempts if collect_rejected_attempts else accepted_solutions_count
        if collected_solutions_count < overlap.min_overlap:
            overlap_increase = overlap.min_overlap - collected_solutions_count
        elif collected_solutions_count == overlap.max_overlap:
            overlap_increase = 0
        else:
            if len(correct_control_tasks_in_accepted) == 0:
                confidence = 0.0
            else:
                confidence = correct_control_tasks_in_accepted.mean() / control_task_count
            if confidence >= needed_confidence:
                overlap_increase = 0
            else:
                overlap_increase = 1
    confidence = (
        correct_control_tasks_in_accepted.mean() / control_task_count
        if len(correct_control_tasks_in_accepted) > 0
        else 0.0
    )
    return (
        loop_iterations,
        accepted_solutions_count,
        attempts,
        hmean([confidence, len(correct_control_tasks_in_accepted) / 4]),
        ok,
    )


def simulate_avg(
    overlap: classification_loop.Overlap,
    worker_q: float,
    control_task_count: int,
    control_tasks_for_accept: int,
    n_iter: int = 100,
    time_budget: float = 3.0,
    collect_rejected_attempts: bool = False,
    accuracy_threshold: Optional[int] = None,
) -> Tuple[float, float, float, float, bool]:
    start = time.time()
    correct_control_tasks_in_solved = np.random.binomial(
        control_task_count, worker_q, (n_iter, control_task_count * 100)
    )
    loop_iterations_sum, attempts_sum, overlap_sum, confidence_sum = 0, 0, 0, 0.0
    ok = True

    for i in range(n_iter):
        current = time.time()
        budget = time_budget - (current - start)
        if budget < 0:
            ok = False
            break
        loop_iterations, iteration_overlap, attempts, confidence, ok = simulate(
            overlap,
            worker_q,
            control_task_count,
            control_tasks_for_accept,
            correct_control_tasks_in_solved=correct_control_tasks_in_solved[i],
            time_budget=budget,
            collect_rejected_attempts=collect_rejected_attempts,
            accuracy_threshold=accuracy_threshold,
        )
        loop_iterations_sum += loop_iterations
        attempts_sum += attempts
        confidence_sum += confidence
        overlap_sum += iteration_overlap
        if not ok:
            break
    return loop_iterations_sum / n_iter, overlap_sum / n_iter, attempts_sum / n_iter, confidence_sum / n_iter, ok


def _get_stats_and_chars(
    task_duration_hint: timedelta,
    task_count: int,
    control_task_count: int,
    assignment_price: float,
    overlap: str,
    overlap_confidence: float,
    control_tasks_for_accept: int,
    num_classes: int,
    verified_lang: bool,
    training: bool,
    training_score: int,
) -> Tuple[List[float], List[str]]:
    sz = 1000
    duration = task_duration_hint * (task_count + control_task_count)
    outputs = []

    commission = pricing.toloka_commission * assignment_price
    price_per_hour = timedelta(hours=1) / duration * assignment_price
    pricing_config = pricing.PoolPricingConfig(
        assignment_price=assignment_price,
        real_tasks_count=task_count,
        control_tasks_count=control_task_count,
    )
    calc_overlap = get_overlap(overlap, overlap_confidence)
    pool_price = client.get_pool_price(sz, pricing_config, overlap=calc_overlap).replace('$', '\\$')
    robustness = 1.0 - _get_random_chance(control_task_count, control_tasks_for_accept, num_classes)

    max_overlap = classification_loop.StaticOverlap(5)
    _, max_price, _ = pricing._get_assignment_params(
        task_count=2,
        task_duration_hint=task_duration_hint,
        price_per_hour=2.0,
    )
    max_pricing_config = pricing.PoolPricingConfig(
        assignment_price=max_price,
        real_tasks_count=1,
        control_tasks_count=1,
    )

    max_pool_price = float(client.get_pool_price(sz, max_pricing_config, overlap=max_overlap)[:-1])

    outputs.append(f'Robustness: {robustness:.2f}, recommended range - [0.6, 0.9]')
    outputs.append(f'Assignment duration: {duration}, recommended range - [30s, 5m]')
    if commission >= pricing.min_toloka_commission:
        outputs.append(f'Assignment commission, usd: {commission:.3f}')
    else:
        outputs.append(
            f'Warning: unprofitable assignment comission, usd: {pricing.min_toloka_commission:.3f} &gt; {commission:.3f}'
        )
    outputs.append(f'Expected worker earnings per hour: {price_per_hour:.3f}\\$')
    outputs.append(f'Expected price per {sz} objects: {pool_price}')

    worker_q = min(
        0.95,
        0.65 + 0.1 * (int(verified_lang) + int(training)) + 0.1 * (0.0 if not training else training_score / 100),
    )
    base_overlap = classification_loop.StaticOverlap(1)
    avg_iter, avg_overlap, _, avg_conf, ok = simulate_avg(
        calc_overlap, worker_q, max(1, control_task_count), control_tasks_for_accept, n_iter=ITERATIONS_COUNT
    )
    if control_task_count == 0:
        avg_conf = 0.1
    if not ok:
        outputs = ["<font color='red'> <b>WARNING: loop may not converge; unreliable results</b>"] + outputs

    pricing_config = pricing.PoolPricingConfig(
        assignment_price=assignment_price,
        real_tasks_count=task_count,
        control_tasks_count=control_task_count,
    )
    base_pool_price = float(client.get_pool_price(sz, pricing_config, base_overlap)[:-1])
    avg_pool_price = base_pool_price * avg_overlap
    outputs.append(f'Expected loop iterations: {avg_iter}')
    if isinstance(calc_overlap, classification_loop.DynamicOverlap):
        outputs.append(f'Expected average overlap: {avg_overlap}')
        outputs.append(f'Expected average price per {sz} objects: {avg_pool_price:.2f}\\$')
    #   outputs.append(f'MAX price per {sz} objects: {max_pool_price}')

    worker_count = CLASSIFICATION_WORKERS_COUNT
    worker_count -= int(task_count + control_task_count > 15)
    worker_count -= int(duration > timedelta(minutes=5))
    worker_count -= int(training)
    worker_count -= int(verified_lang)
    worker_count -= int(False if control_task_count == 0 else control_tasks_for_accept / control_task_count > 0.8)
    worker_count -= int(price_per_hour < 0.3)
    worker_count += int(price_per_hour > 0.9)

    outputs.append(f'Task attractiveness index: {worker_count - 1} / {CLASSIFICATION_WORKERS_COUNT}')

    c = avg_pool_price / max_pool_price * 3
    q = hmean([avg_conf, robustness])
    s = (1.0 - avg_iter / 5) * worker_count / CLASSIFICATION_WORKERS_COUNT
    t = 1.0 - s

    return [q, t, c], outputs


def _get_annotation_stats_and_chars(
    task_duration_hint: timedelta,
    task_count: int,
    assignment_price: float,
    check_sample: bool,
    max_tasks_to_check: int,
    accuracy_threshold: int,
    overlap: str,
    overlap_confidence: float,
    control_tasks_for_accept: int,
    verified_lang: bool,
) -> Tuple[List[float], List[str]]:
    sz = 1000
    duration = task_duration_hint * task_count
    outputs = []

    commission = pricing.toloka_commission * assignment_price
    price_per_hour = timedelta(hours=1) / duration * assignment_price
    calc_overlap = classification_loop.DynamicOverlap(
        min_overlap=1, max_overlap=int(overlap), confidence=overlap_confidence
    )

    max_task_count = task_count
    if check_sample:
        max_task_count = max_tasks_to_check

    # todo: is static overlap as max measure ok?
    max_overlap = classification_loop.StaticOverlap(5)
    _, max_price, _ = pricing._get_assignment_params(
        task_count=1,
        task_duration_hint=task_duration_hint,
        price_per_hour=2.0,
    )
    max_pricing_config = pricing.PoolPricingConfig(
        assignment_price=max_price,
        real_tasks_count=1,
        control_tasks_count=0,
    )
    # todo: we need to use get_annotation_price here with the clf pricing config
    max_pool_price = float(client.get_pool_price(sz, max_pricing_config, overlap=max_overlap)[:-1])

    outputs.append(f'Assignment duration: {duration}, recommended range - [30s, 5m]')
    if commission >= pricing.min_toloka_commission:
        outputs.append(f'Assignment commission, usd: {commission:.3f}')
    else:
        outputs.append(
            f'Warning: unprofitable assignment comission, usd: {pricing.min_toloka_commission:.3f} &gt; {commission:.3f}'
        )
    outputs.append(f'Expected worker earnings per hour: {price_per_hour:.3f}\\$')

    worker_q = min(0.95, 0.65 + 0.1 * int(verified_lang))

    base_overlap = classification_loop.StaticOverlap(1)
    avg_iter, avg_accepted, avg_attempts, avg_conf, ok = simulate_avg(
        calc_overlap,
        worker_q,
        max(1, max_task_count),
        control_tasks_for_accept,
        n_iter=ITERATIONS_COUNT,
        collect_rejected_attempts=True,
        accuracy_threshold=accuracy_threshold if check_sample else None,
    )

    if not ok:
        outputs = ["<font color='red'> <b>WARNING: loop may not converge; unreliable results</b>"] + outputs

    pricing_config = pricing.PoolPricingConfig(
        assignment_price=assignment_price,
        real_tasks_count=task_count,
        control_tasks_count=0,
    )
    base_pool_price = float(client.get_pool_price(sz, pricing_config, base_overlap)[:-1])
    avg_pool_price = base_pool_price * avg_accepted
    outputs.append(f'Expected loop iterations: {avg_iter}')
    outputs.append(f'Expected average overlap: {avg_attempts}')

    worker_count = ANNOTATION_WORKERS_COUNT
    worker_count -= int(task_count > 15)
    worker_count -= int(duration > timedelta(minutes=5))
    worker_count -= int(verified_lang)
    worker_count -= int(control_tasks_for_accept / max_tasks_to_check > 0.8)
    worker_count -= int(price_per_hour < 0.3)
    worker_count += int(price_per_hour > 0.9)

    outputs.append(f'Task attractiveness index: {worker_count - 1} / {ANNOTATION_WORKERS_COUNT}')

    c = avg_pool_price / max_pool_price * 5
    q = avg_conf
    s = (1.0 - (avg_iter + avg_attempts) / 10) * worker_count / ANNOTATION_WORKERS_COUNT
    t = 1.0 - s

    return [q, t, c], outputs


def _get_cumulative_stats_and_characteristics(
    check_quality: float,
    check_time: float,
    check_cost: float,
    markup_quality: float,
    markup_time: float,
    markup_cost: float,
    check_task_count: int,
    check_control_task_count: int,
    check_assignment_price: float,
    markup_task_count: int,
    markup_assignment_price: float,
    check_overlap: str,
    check_overlap_confidence: float,
    markup_overlap: str,
    markup_overlap_confidence: float,
    markup_check_sample: bool,
    markup_max_tasks_to_check: int,
    markup_accuracy_threshold: int,
) -> Tuple[List[Union[float, Tuple[float, str]]], List[str]]:
    sz = 1000

    check_pricing_config = pricing.PoolPricingConfig(
        real_tasks_count=check_task_count,
        control_tasks_count=check_control_task_count,
        assignment_price=check_assignment_price,
    )

    markup_pricing_config = pricing.PoolPricingConfig(
        real_tasks_count=markup_task_count,
        control_tasks_count=0,
        assignment_price=markup_assignment_price,
    )
    assignment_check_sample = None
    extra_text = ''
    check_coverage = 1.0
    if markup_check_sample:
        assignment_check_sample = evaluation.AssignmentCheckSample(
            max_tasks_to_check=markup_max_tasks_to_check,
            assignment_accuracy_finalization_threshold=markup_accuracy_threshold,
        )
        check_coverage = markup_max_tasks_to_check / markup_task_count
        color = 'black'
        if check_coverage < 0.5:
            color = 'red'
        extra_text = f"<font color='{color}'>Check coverage: {check_coverage:.2f}</font>"

    pool_price_formula = client.AnnotationPriceFormula(
        input_objects_count=sz,
        markup_config=markup_pricing_config,
        check_config=check_pricing_config,
        markup_overlap=get_overlap(f'1-{markup_overlap}', markup_overlap_confidence),
        check_overlap=get_overlap(check_overlap, check_overlap_confidence),
        assignment_check_sample=assignment_check_sample,
    )

    return [
        (hmean([check_quality, markup_quality]), extra_text),
        (check_time + check_coverage * markup_time) / 2,
        (check_cost + check_coverage * markup_cost) / 2,
    ], [f'Expected price for {sz} objects:'] + [f'$${v}$$' for v in pool_price_formula.clear_formula()]
